from typing import List
from pydantic import BaseModel
from nazca4sdk.datahandling.npoe.meter import Meter


class Connection(BaseModel):
    systemName: str
    organizationId: str
    id: str
    orderedPower: int
    meters: List[Meter]


class Connections(BaseModel):
    __root__: List[Connection]

    def connections(self):
        return self.__root__



"""Module for forecasting"""
from nazca4sdk.analytics.prediction.predictions import Predictions, forecast_prediction
from nazca4sdk.datahandling.variable_verificator import VariableIntervalSubtractionInfo, VariableType, \
    ForecastForPrediction
from nazca4sdk.system.system_cache import SystemCache
from pydantic import ValidationError


class Forecasting:
    """
    Forecasting module with functions for forecasting
    """

    def __init__(self, cache: SystemCache):
        self._system_cache = cache

    def predict(self, module_name: str, variable_names: str, time_amount: int, time_unit: str,
                forecast_time_amount: int, forecast_time_unit: str, prediction_tool: str = "prophet"):
        """
        Predict value for the future based on prophet procedure

        Args:
            module_name: name of module,
            variable_names: list of variable names,
            time_amount: beginning of the time range
            time_unit: 'DAY','HOUR'...
            forecast_time_amount: int: periods forecast: 10,50,60
            forecast_time_unit: str: frequency for forecast time: '1min', '5min'...
            prediction_tool: str: "prophet" or "nixtla"

        Returns:
            result dict with:
                lower_bound, upper_bound, prediction value, mse error, mae errors
        """

        try:
            exist = self._system_cache.check_if_exist(module_name, [variable_names])
            if not exist:
                print(f"Module {module_name} or {variable_names} not exist")
                return None
            variables_group = self._system_cache.group_variables(module_name, [variable_names])
            variables_grouped = {'variables_grouped': variables_group}
            variable = VariableType(**variables_grouped).variable_verify_type(variables_group)

            input_data = {
                "module_name": module_name,
                "variable_names": [variable],
                "time_amount": time_amount,
                "time_unit": time_unit
            }
            data_input = dict(
                VariableIntervalSubtractionInfo(**{"time_amount": time_amount,
                                                   "time_unit": time_unit
                                                   }))

            input_forecast = {
                "forecast_time_amount": forecast_time_amount,
                "forecast_time_unit": forecast_time_unit,
                "prediction_tool": prediction_tool
            }

            data_forecast = dict(ForecastForPrediction(**input_forecast))

            result = forecast_prediction(Predictions(input_data['module_name'], input_data['variable_names'][0],
                                                     data_input['time_amount'], data_input['time_unit'],
                                                     data_forecast['forecast_time_amount'],
                                                     data_forecast['forecast_time_unit'],
                                                     data_forecast['prediction_tool']))
            return result

        except ValidationError as error:
            print(error.json())
            return None



"""Module for predictions with Prophet/Nixtla"""
from pydantic import ValidationError
from nazca4sdk.analytics.__brain_connection import BrainClient
from dataclasses import dataclass


@dataclass
class Predictions:
    """
    Prediction Class

    Attributes:
        module_name: str
        variable_names: str
        time_amount: int
        time_unit: str
        forecast_time_amount: int
        forecast_time_unit: str
        prediction_tool: str


    Methods:

        forecast_prediction()
            Determine prediction values for determined input
    """
    module_name: str
    variable_names: str
    time_amount: int
    time_unit: str
    forecast_time_amount: int
    forecast_time_unit: str
    prediction_tool: str


def forecast_prediction(prediction_info: Predictions):
    """Function to determine prediction values for determined input

    Args:
        prediction_info: prediction_info
    Returns:
        dict: prediction parameters
    """

    brain = BrainClient()

    try:
        input_data = {
            "module_name": prediction_info.module_name,
            "variable_names": prediction_info.variable_names,
            "time_amount": prediction_info.time_amount,
            "time_unit": prediction_info.time_unit,
            "forecast_time_amount": prediction_info.forecast_time_amount,
            "forecast_time_unit": prediction_info.forecast_time_unit,
            "prediction_tool": prediction_info.prediction_tool
        }

        data = {**input_data}

        response = brain.get_prediction(data)
        result = brain.parse_response(response)

        return result

    except ValidationError as error:
        print(error.json())
        return None



import pandas as pd
from datetime import datetime
from typing import Optional
from pandas import DataFrame
from nazca4sdk.datahandling.hotstorage.clickhouse.clickhouse_client import ClickhouseClient
from nazca4sdk.datahandling.hotstorage.query import Query
from nazca4sdk.datahandling.variable_verificator import VariableIntervalInfo, DATE_TIME_FORMAT, \
    VariableIntervalSubtractionInfo, VariableOverDayValidator
from nazca4sdk.system.system_cache import SystemCache
from nazca4sdk.tools.time import get_time_delta
from nazca4sdk.datahandling.data_mod import Data
from dependency_injector.wiring import inject


class Variables:
    """ Class to get module variable values

    """
    @inject
    def __init__(self, cache: SystemCache, click: ClickhouseClient):
        self._system_cache = cache
        self._clickhouse_client = click

    def list(self):
        """ List all module variables in system

        Returns:
            list: modules list
        """
        return self._system_cache.variables

    def read(self, module_name: str, variable_names: list, **time: dict) -> Optional[DataFrame]:
        """
        Gets module variables values in specific time range

        Args:
            module_name: name of module,
            variable_names: list of variable names,
            time: possible pairs - start_date, end_date or time_amount, time_unit

        Returns:
            DataFrame: values for selected variables and time range

        Examples:
            read(module_name:"module", variable_names:["V1"], start_date="2023-03-21TOO:OO:00")
            read(module_name:"module", variable_names:["V1"],time_amount=1,time_unit="MINUTE")
        """
        keys = time.keys()

        columns_to_validate = {'module_name': module_name,
                               'variable_names': variable_names}
        columns_info = VariableOverDayValidator(**columns_to_validate)

        if "start_date" in keys and "end_date" in keys:
            time_to_validate = {'start_date': time.get("start_date"),
                                'end_date': time.get("end_date")}

            time_info = VariableIntervalInfo(**time_to_validate)
            return self._variable_over_day(columns_info.module_name,
                                           columns_info.variable_names,
                                           datetime.strptime(time_info.start_date, DATE_TIME_FORMAT),
                                           datetime.strptime(time_info.end_date, DATE_TIME_FORMAT))
        elif "time_amount" in keys and "time_unit" in keys:
            time_amount = time.get("time_amount")
            time_unit = time.get("time_unit")

            time_to_validate = {'time_amount': time_amount,
                                'time_unit': time_unit}

            variable_info = VariableIntervalSubtractionInfo(**time_to_validate)

            end_date = datetime.now()
            start_date = end_date - get_time_delta(variable_info.time_unit, variable_info.time_amount)

            return self._variable_over_day(columns_info.module_name,
                                           columns_info.variable_names,
                                           start_date,
                                           end_date)

        print("time should contains start_date and end_date or time_unit and time_amount")
        return None

    def pivot(self, module_name: str, variable_names: list, **time: dict):
        dataframe = self.read(module_name, variable_names, **time)
        return 'Empty dataframe' if dataframe is None else Data(dataframe).to_pivot_df()

    def stats(self, module: str, variables: [str], start_date: str, end_date: str) -> Optional[DataFrame]:
        """ Read module variables stats

        Args:
            module: module name
            variables: list of variable names
            start_date: start of date range
            end_date: end of date range
        Returns:
            DataFrame:  Variables Stats
        """
        return self._read_variables_stats(module=module, variables=variables, start_date=start_date, end_date=end_date)

    def _read_variables_stats(self, module: str, variables: [str], start_date: str, end_date: str) \
            -> Optional[DataFrame]:
        exist_vars = self._system_cache.check_if_exist(module, variables)
        if not exist_vars:
            return None
        variables_grouped = self._system_cache.group_variables(module, variables)
        # check variables type
        for element in variables_grouped:
            if element[0] not in self._system_cache.stats_variable_type:
                print("Variable to count stats should be type of:")
                for element_type in self._system_cache.stats_variable_type:
                    print(f" -  {element_type}")
                print(f"Variables {element[1]} is type of {element[0]}")
                return None

        dataframe = pd.DataFrame()
        for key, value in variables_grouped:
            variables_list = ",".join([f"'{item}'" for item in value])
            query = Query() \
                .SELECT("Module, Variable, toFloat32(min(Value)) as Min, "
                        "toFloat32(max(Value)) as Max, toFloat64(avg(Value)) as Avg, "
                        "toFloat32(any(Value)) as FirstValue, "
                        "toFloat32(anyLast(Value)) as LastValue, toFloat64(varPop(Value)) as Variance, "
                        "toFloat64(stddevPop(Value)) as Std ") \
                .FROM(f"nazca.devices_data_{key}") \
                .WHERE(f"Module like '{module}' and "
                       f"MeasureTime >= '{start_date}' "
                       f"and MeasureTime <= '{end_date}' "
                       f"and Variable IN ({variables_list})") \
                .GROUP_BY("Module, Variable")
            df = self._clickhouse_client.get(query)
            dataframe = pd.concat([dataframe, df])
        return dataframe

    def _variable_over_day(self, module_name, variable_names, start_date, end_date) -> Optional[DataFrame]:
        """
        Gets variable in specific time range by connection with open database

        Args:
            module_name - name of module,
            variable_names - list of variable names,
            start_time - beginning of the time range
            stop_time - ending of the time range

        Returns:
            DataFrame: values for selected variable and time range

        """

        try:
            exist_vars = self._system_cache.check_if_exist(module_name, variable_names)
            if not exist_vars:
                return None
            variables_grouped = self._system_cache.group_variables(
                module_name, variable_names)

            dataframe = pd.DataFrame()
            for group in variables_grouped:
                table = group[0]
                variables = ",".join([f"'{item}'" for item in group[1]])
                query = Query() \
                    .SELECT("*") \
                    .FROM(f"nazca.devices_data_{table}") \
                    .WHERE(f"Module like '{module_name}' "
                           f"and MeasureTime >= '{start_date.strftime('%Y-%m-%dZ%H:%M:%S')}' "
                           f"and MeasureTime <= '{end_date.strftime('%Y-%m-%dZ%H:%M:%S')}' "
                           f"and Variable IN ({variables})")
                df = self._clickhouse_client.get(query)
                try:
                    dataframe = pd.concat([dataframe, df]).sort_values('MeasureTime').reset_index(drop=True)
                except:
                    raise ValueError("No data")
            return dataframe
        except ValueError:
            print("Error - Get variable data from Nazca4")
            return None



"""Set of classes to prepare objects for knowledge"""
import json


class FormatText():
    """Class to prepare object of Format KnowledgeDataType"""

    def __init__(self, value, **kwargs):
        """ Initializing the FormatText object

        Args:
            value: value to write to knowledge,
            **kwargs: oprional keyword arguments for Format KnowledgeDataType:
                pipe: Angular pipe name,
                args: Angular pipe parameters,
                prefix: prefix,
                suffix: suffix
                color: name of color

        Returns:
            Object to use for Format KnowledgeDataType

        Example:
            FormatText('1995-12-17T03:24:00', pipe='DatePipe', args='fullDate', prefix='Dzisiaj jest', suffix='dzien', color='red')
        """

        self.kwargs = kwargs
        self.kwargs['value'] = value

    def __str__(self):
        """ Change to proper string """

        return str(json.dumps(self.kwargs))


class FormatChart():
    """Class to prepare object of chart KnowledgeDataType"""

    def __init__(self, chart_type, module, variable, start_time, stop_time):
        """ Initializing the FormatText object

        Args:
            chart_type: type of chart to create,
            module: name of module,
            variable: name of variable to plot,
            start_time: start time of chart range,
            stop_time: stop time of chart range,
            source: name of table in ClickHouse,

        Returns:
            Object to use for Chart KnowledgeDataType

        Example:
            FormatChart('timeseries', 'Socomec', 'P', '2021-12-17T03:00:00Z', '2021-12-17T05:00:00Z')
        """
        self.chartType = chart_type
        self.properties = {
            "module": module,
            "variable": variable,
            "startTime": start_time,
            "endTime": stop_time}

    def __str__(self):
        """ Change to proper string """

        return str(json.dumps(self.__dict__))



"""Cache module"""
import json

import requests

from nazca4sdk.datahandling.open_data_client import OpenDataClient
from nazca4sdk.datahandling.variable_historical_info import VariableHistoricalInfo
from dependency_injector.wiring import inject

JSON_HEADERS = {'Content-Type': 'application/json',
                'Accept': 'application/json'}


class SystemCache:
    """
    System caching module to as a second layer of opendata with SDK
    """

    types_dict = {"IntegerT": "int", "Float32T": "float", "TimeT": "datetime",
                  "BooleanT": "bool", "StringT": "string", "UIntegerT": "int"}
    stats_variable_type = {"float", "int"}

    @inject
    def __init__(self, opendata_client: OpenDataClient):
        """
        Initializing opendata with HotStorage to receive system configuration

        """

        self._opendata = opendata_client
        self._base_url = self._opendata.base_url
        self._types = {}
        self.modules = []
        self.variables = {}

    @property
    def load(self):
        """
        Loading definitions of modules and variables

        Returns:
            list: module_name list
            list: variable_name list

        """
        try:
            # TODO przeniesc do OpenDataClient wywolanie a tu tylko parsowac, usuanac wtedy base_url jak public
            response = requests.get(
                f'{self._base_url}/api/Config/modulesDefinitions', verify=False)
            if response.status_code == 200:
                json_response = response.json()
                for element in json_response:
                    module = element['identifier']
                    self.modules.append(module)
                    definition_string = element['definition']
                    definition = json.loads(definition_string)
                    variables = definition["variables"]
                    variable_list = []
                    var_dict = {}
                    for variable in variables:
                        name = variable["name"]
                        variable_type = variable["type"]
                        variable_list.append(name)
                        try:
                            readable_variable_type = self.types_dict[variable_type]
                            var_dict[name] = readable_variable_type
                        except KeyError:
                            print(
                                f"Module {module} Variable {name} type {variable_type} not recognized!")
                    self.variables[module] = variable_list
                    self._types[module] = var_dict
                return True
            return False
        except requests.exceptions.ConnectionError:
            print("A Connection error occurred.")
            return False

    def read_historical_variable(self, variable_historical_info: VariableHistoricalInfo):
        """
        Get paged variables in specific time range

        Args:
            variable_historical_info: info about variable

        Returns:
            array of variable values : Paged Variable values from selected time range
        Example:
            read_historical_variable(variable_historical_info)
        """

        try:
            exist_vars = self.check_if_exist(variable_historical_info.module_name,
                                             variable_historical_info.variable_names)
            if not exist_vars:
                print(f"Module {variable_historical_info.module_name} or "
                      f"{variable_historical_info.variable_names} not exist")
                return None
            variables_grouped = self.group_variables(variable_historical_info.module_name,
                                                     variable_historical_info.variable_names)
            params = [
                ('module', variable_historical_info.module_name),
                ('startdate', variable_historical_info.start_date),
                ('enddate', variable_historical_info.end_date)
            ]

            for variable_type in variables_grouped:
                for variable_name in variable_type[1]:
                    params.append(
                        (f'groupedVariables[{variable_type[0]}]', variable_name))
            response = self._opendata.get_paged_data(url="/api/hotstorage/ReadHistoricalVariable",
                                                     params=params, page_size=variable_historical_info.page_size)
            return response
        except requests.exceptions.ConnectionError:
            print("Error - Get data from OpenData")
            return None

    def group_variables(self, module: str, variable: list):
        """Group variables by type

        Args:
            module: module
            variable: list of variables

        Returns:
            List of tuples with variables grouped by type:
                [(type,[variables])]

        """

        grouped_variables = list(
            map(lambda x: [x, self._types[module][x]], variable))
        tables = set(map(lambda x: x[1], grouped_variables))
        variables_grouped = [
            (x, [y[0] for y in grouped_variables if y[1] == x]) for x in tables]
        return variables_grouped

    def check_if_exist(self, module: str, variables: list):
        """Verify if module or variable are in system

        Args:
            module:module
            variables:list of variables

        Returns:
            True(bool) if exists
            raise ArgumentMissing error if not exists

        """

        if module not in self._types:
            print(f'Module: {module} not found')
            return False
        for variable in variables:
            if variable not in self._types[module]:
                print(f'Variable: {variable} not found')
                return False

        return True



"""Module to calculate vibration quality"""
import logging
from pydantic import ValidationError
from nazca4sdk.analytics.__brain_connection import BrainClient


class VibrationQuality:
    """ Class to send and receive information from Vibration quality module in brain"""

    def __init__(self):
        self.vibration_brain = BrainClient()

    def calculate_vibration_quality(self, group: str, module: str):
        """Function to determine vibration quality values for determined input

        Args:
            group: group name
            module: module name

        Returns:
            dict: vibration quality parameters
        """

        try:
            logging.info(f'Module Name: {module}, Group: {group}')
            response = self.vibration_brain.get_vibration_quality(group, module)
            result = self.vibration_brain.parse_response(response)
            logging.info(f'Response: {result}')
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def calculate_vibration_quality_by_params(self, group: str, vibration):
        """Function to determine vibration quality values for determined input

        Args:
            group: group name
            vibration: vibration value

        Returns:
            vibration quality parameters based on ISO norm
        """

        try:
            logging.info(f'Vibration: {vibration}, Group: {group}')

            if not isinstance(vibration, float):
                return TypeError(f'vibration is not a valid float, should be {float(vibration)}')
            response = self.vibration_brain.get_vibration_quality_by_params(group, vibration)
            result = self.vibration_brain.parse_response(response)
            logging.info(f'Response: {result}')
            return result
        except ValidationError as error:
            print(error.json())
            return None



"""Testing nazca4sdk """
from nazca4sdk import SDK
from nazca4sdk.datahandling.file.file_storage import FileStorage

sdk = SDK()
# print(sdk.modules)


# file = sdk.file_storage.download_file("/siemka2.txt", "./siemka.txt")
sdk.file_storage.send_file('/Sodomita.txt', "./siemka.txt")
print("OK")



from nazca4sdk.datahandling.hotstorage.helper import transform, create_variables_frame
from nazca4sdk.datahandling.hotstorage.model.user_variable import UserVariable
from nazca4sdk.datahandling.kafka.kafka_sender import KafkaSender
from dependency_injector.wiring import inject


class UserVariableSaver:
    @inject
    def __init__(self, kafka_client: KafkaSender):
        self._kafka_sender = kafka_client

    def save_variables(self, user_variables: [UserVariable]) -> bool:
        variables = transform(user_variables)
        if len(variables) == 0:
            return False
        user_variables_frame = create_variables_frame(variables)
        return self._kafka_sender.send_message("dataflow.fct.clickhouse", 'Variables', user_variables_frame)



from datetime import datetime

from pydantic import BaseModel, validator

types = ["bool", "int", "double", "datetime", "text"]
DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
DATE_TIME_EXAMPLE = "yyyy-mm-ddTHH:MM:SS"


class UserVariableInfoValue(BaseModel):
    type: str
    name: str

    @validator("type")
    def validate_type(cls, valid_type):
        if valid_type not in types:
            raise ValueError(f"type should be one of {types}")
        return valid_type


class UserVariablesInfoValues(BaseModel):
    variables = {}
    startDate: str
    endDate: str

    @validator("endDate")
    def validate_end_date(cls, valid_datetime):
        """Validator to check if time amount is correct

               Args:
                   valid_datetime: str

               Returns:
                   valid_datetime: str
               """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                f'Bad datetime format. Example {DATE_TIME_EXAMPLE} ')

    @validator("startDate")
    def valid_start_date(cls, valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                f'Bad datetime format. Example {DATE_TIME_EXAMPLE}')

    def add_variable(self, variable: UserVariableInfoValue):

        if variable.type not in self.variables:
            self.variables[variable.type] = []
        self.variables[variable.type].append(variable.name)

    @staticmethod
    def valid_datetime(valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                f'Bad datetime format. Format {DATE_TIME_EXAMPLE}')



import pandas as pd
from pandas import DataFrame
from typing import Optional
from nazca4sdk import UserVariableInfo, UserVariable, UserVariableDataType
from nazca4sdk.datahandling.hotstorage.clickhouse.clickhouse_client import ClickhouseClient
from nazca4sdk.datahandling.hotstorage.helper import userVariableTableNames
from nazca4sdk.datahandling.hotstorage.query import Query
from nazca4sdk.datahandling.hotstorage.user_variable_saver import UserVariableSaver
from dependency_injector.wiring import inject


class UserVariables:
    @inject
    def __init__(self, click: ClickhouseClient, user_variable_saver: UserVariableSaver):
        self._clickhouse_client = click
        self._user_variables_saver = user_variable_saver

    def read(self, variables: [UserVariableInfo], start_date: str, end_date: str) -> DataFrame:
        """Read user variables from hot storage

               Args:
                   start_date: begin of time range
                   end_date: end of time range
                   variables: list of user variables to read

               Returns:
                   DataFrame with variables values
               """
        grouped_variables = dict()
        for d in variables:
            variables_type = userVariableTableNames[d.type]
            if variables_type not in grouped_variables.keys():
                grouped_variables[variables_type] = [d.name]
            else:
                grouped_variables[variables_type].append(d.name)

        dataframe = pd.DataFrame()
        for table in grouped_variables:
            variables_list = ",".join([f"'{item}'" for item in grouped_variables[table]])
            query = Query() \
                .SELECT("*") \
                .FROM(f"nazca.user_data_{table}") \
                .WHERE(f"DateTime >= '{start_date}' "
                       f"and DateTime <= '{end_date}' "
                       f"and Variable IN ({variables_list})")
            df = self._clickhouse_client.get(query)
            dataframe = pd.concat([dataframe, df])
        return dataframe

    def write(self,  user_variables: [UserVariable]) -> bool:
        """
        Save user variables in hot storage

        Args:
            user_variables: list of user variables to save

        Returns:
            True - variables to save send to hot storage
            False - communication with hot storage error
        """

        return self._user_variables_saver.save_variables(user_variables)

    def stats(self, variables: [UserVariableInfo], start_date: str, end_date: str) -> Optional[DataFrame]:
        """Read user variables stats

           Args:
               variables: list of variables names
               start_date: start of date range
               end_date: end of date range
           Returns:
               VariableStats DataFrame
        """

        if not isinstance(variables, list):
            print("variables should be list")
            return None
        grouped_variables = dict()
        for d in variables:
            if d.type == UserVariableDataType.TEXT or d.type == UserVariableDataType.BOOL:
                print(f"user_variables_stats only for INT and DOUBLE type, not for {d.type}")
                continue
            variables_type = userVariableTableNames[d.type]
            if variables_type not in grouped_variables.keys():
                grouped_variables[variables_type] = [d.name]
            else:
                grouped_variables[variables_type].append(d.name)

        dataframe = pd.DataFrame()
        for table in grouped_variables:
            variables_list = ",".join([f"'{item}'" for item in grouped_variables[table]])
            query = Query() \
                .SELECT("Variable, toFloat32(min(Value)) as Min, "
                        "toFloat32(max(Value)) as Max, toFloat64(avg(Value)) as Avg, "
                        "toFloat32(any(Value)) as FirstValue, "
                        "toFloat32(anyLast(Value)) as LastValue, toFloat64(varPop(Value)) as Variance, "
                        "toFloat64(stddevPop(Value)) as Std ") \
                .FROM(f"nazca.user_data_{table}") \
                .WHERE(f"DateTime >= '{start_date}' "
                       f"and DateTime <= '{end_date}' "
                       f"and Variable IN ({variables_list})") \
                .GROUP_BY("Variable")
            df = self._clickhouse_client.get(query)
            dataframe = pd.concat([dataframe, df])
        return dataframe



from nazca4sdk.sdk import SDK

sdk = SDK(False)
print(sdk.variables.read('symulator', ['V1'], time_unit='HOUR', time_amount=5))

print(sdk.variables.pivot('symulator', ['V1', 'V2'], time_unit='HOUR', time_amount=5))



from nazca4sdk.datahandling.hotstorage.model.user_variable import UserVariable, UserVariableInfo, UserVariableDataType
from nazca4sdk.sdk import SDK
from nazca4sdk.datahandling.knowledge.knowledge_data_type import KnowledgeDataType
from nazca4sdk.datahandling.knowledge.knowledge_objects import FormatText, FormatChart


from enum import Enum


class KnowledgeDataType(str, Enum):
    TEXT = "text"
    NUM = "num"
    FORMAT = "format"
    IMAGE = "img"
    BLOB = "blob"
    LINK = "link"
    BOOL = "bool"
    CHART = "chart"

    @classmethod
    def has_value(cls, value):
        if not isinstance(value, str):
            return False
        return value in cls._value2member_map_



from pydantic import ValidationError

from nazca4sdk.datahandling.cache.key_value import KeyValue
from nazca4sdk.datahandling.open_data_client import OpenDataClient
from dependency_injector.wiring import inject


class CacheStorage:
    """Allow user to read and write value to Cache

    """
    @inject
    def __init__(self, opendata_client: OpenDataClient):
        self._opendata = opendata_client

    def read_keys(self, keys):
        """Read cache values for list of keys
        
        Args:
            keys - list of keys to read
        Returns:
            List of cache entry
        """
        return self._opendata.read_cache_keys(keys)

    def write_keys(self, key: str, value):
        """write key value to cache

        Args:
            key: key
            value: value to write to cache
        Returns:
            CacheEntry if success or None if error
        """

        data = {'key': key,
                'value': value}
        try:
            variable_info = KeyValue(**data)
            return self._opendata.write_cache_keys(variable_info)
        except ValidationError as error:
            print(error.json())
            return None



from nazca4sdk.sdk import SDK

sdk = SDK(False)
sdk.modules
# print(sdk.read_by_query(query='SELECT * FROM nazca.devices_data_int LIMIT 200'))
print(sdk.read_by_query(query='SELECT Module, MeasureTimeLong, MeasureTime, MeasureDate, Variable, Value FROM nazca.devices_data_int order by MeasureTimeLong desc LIMIT 100;'))
# print(list('SELECT * FROM nazca.test LIMIT 5'))



import os
from dependency_injector import containers, providers

from nazca4sdk.analytics.analytics import Analytics
from nazca4sdk.datahandling.cache.cache_storage import CacheStorage
from nazca4sdk.datahandling.file.file_storage import FileStorage
from nazca4sdk.datahandling.hotstorage.clickhouse.clickhouse_client import ClickhouseClient
from nazca4sdk.datahandling.hotstorage.user_variable_saver import UserVariableSaver
from nazca4sdk.datahandling.kafka.kafka_sender import KafkaSender
from nazca4sdk.datahandling.knowledge.knowledge_storage import KnowledgeStorage
from nazca4sdk.datahandling.nazcavariables.nazca_variables_storage import NazcaVariablesStorage
from nazca4sdk.datahandling.open_data_client import OpenDataClient
from nazca4sdk.system.npoe import Npoe
from nazca4sdk.system.system_cache import SystemCache
from nazca4sdk.system.user_variables import UserVariables
from nazca4sdk.system.variables import Variables


class Container(containers.DeclarativeContainer):

    config = providers.Configuration()

    # konfiguracja zmiennych srodowiskowych
    os.environ["https"] = "False"
    # os.environ["opendata_url"] = "opendata"
    os.environ["opendata_url"] = "10.217.81.75"
    # os.environ["clickhouse_host"] = "click"
    os.environ["clickhouse_host"] = "10.217.81.75"
    # os.environ["clickhouse_port"] = "8000"
    os.environ["clickhouse_port"] = "10800"
    # os.environ["kafka_host"] = "broker"
    os.environ["kafka_host"] = "10.217.81.75"
    # os.environ["kafka_port"] = "9092"
    os.environ["kafka_port"] = "10092"
    # zapis konfiguracji ze zmiennych srodowiskowych
    config.https.from_env("https")
    config.opendata_url.from_env("opendata_url")
    config.clickhouse_host.from_env("clickhouse_host")
    config.clickhouse_port.from_env("clickhouse_port")
    config.kafka_host.from_env("kafka_host")
    config.kafka_port.from_env("kafka_port")
    # singleton
    clickhouse_client = providers.Singleton(ClickhouseClient,
                                            host=config.clickhouse_host,
                                            port=config.clickhouse_port)
    opendata_client = providers.Singleton(OpenDataClient,
                                          https=config.https,
                                          url=config.opendata_url)
    kafka_client = providers.Singleton(KafkaSender, host=config.kafka_host, port=config.kafka_port)
    system_cache = providers.Singleton(SystemCache, opendata_client=opendata_client)
    # fabryki
    cache_storage = providers.Factory(CacheStorage, opendata_client=opendata_client)
    nazca_variable_storage = providers.Factory(NazcaVariablesStorage, opendata_client=opendata_client)
    knowledge_storage = providers.Factory(KnowledgeStorage,
                                          opendata_client=opendata_client,
                                          kafka_client=kafka_client)
    npoe = providers.Factory(Npoe, opendata_client=opendata_client, cache=system_cache, click=clickhouse_client,
                             broker=kafka_client)
    user_variable_saver = providers.Factory(UserVariableSaver, kafka_client=kafka_client)
    user_variables = providers.Factory(UserVariables,
                                       click=clickhouse_client,
                                       user_variable_saver=user_variable_saver)
    variables = providers.Factory(Variables, cache=system_cache, click=clickhouse_client)
    analytics = providers.Factory(Analytics, cache=system_cache)
    file_storage = providers.Factory(FileStorage, opendata_client=opendata_client)



from datetime import timedelta


def get_time_delta(time_unit, time_amount):
    if time_unit == 'SECOND':
        return timedelta(seconds=time_amount)
    if time_unit == 'MINUTE':
        return timedelta(minutes=time_amount)
    if time_unit == 'HOUR':
        return timedelta(hours=time_amount)
    if time_unit == 'DAY':
        return timedelta(days=time_amount)
    if time_unit == 'WEEK':
        return timedelta(weeks=time_amount)
    if time_unit == 'YEAR':
        return timedelta(days=time_amount * 365)
    return None



from datetime import datetime
from nazca4sdk.datahandling.kafka.kafka_sender import KafkaSender
from nazca4sdk.datahandling.knowledge.knowledge_data_type import KnowledgeDataType
from nazca4sdk.datahandling.open_data_client import OpenDataClient
from datetime import timezone
from dependency_injector.wiring import inject


class KnowledgeStorage:

    @inject
    def __init__(self, opendata_client: OpenDataClient, kafka_client: KafkaSender):
        self._opendata = opendata_client
        self._kafka_client = kafka_client

    def read_keys(self, name: str = None, ts_min: str = None, ts_max: str = None, size: int = 0) -> list:
        """ read knowledge keys

            Args:
               name: key filter
               ts_min: timestamp min
               ts_max: timestamp max
               size: max key list size
            Returns:
               key list
        """

        def validate(date):
            try:
                if date is None:
                    return True
                datetime.strptime(date, '%Y-%m-%dT%H:%M:%S')
                return True
            except ValueError:
                print("Incorrect datetime format, should be YYYY-MM-DDTHH:MM:SS")
                return False

        if not validate(ts_min):
            return list()
        if not validate(ts_max):
            return list()

        if ts_min is None:
            ts_min_str = None
        else:
            dt_ts_min = datetime.strptime(ts_min, '%Y-%m-%dT%H:%M:%S')
            ts_min_str = dt_ts_min.strftime("%Y%m%dT%H%M%S")

        if ts_max is None:
            ts_max_str = None
        else:
            dt_ts_max = datetime.strptime(ts_max, '%Y-%m-%dT%H:%M:%S')
            ts_max_str = dt_ts_max.strftime("%Y%m%dT%H%M%S")

        return self._opendata.read_knowledge_keys(name, ts_min_str, ts_max_str, size)

    def read_key_values(self, key: str) -> list:
        """ read knowledge for key

               Args:
                   key: key
               Returns:
                   Knowledge
        """
        return self._opendata.read_knowledge_values(key)

    def write_key_values(self, key: str, section: str, value: str, datatype: KnowledgeDataType) -> bool:
        """Write value to knowledge

               Args:
                   key: key
                   section: section name
                   value: value
                   datatype: data type to write
               Returns: True - write success, False - write error
               """

        if not KnowledgeDataType.has_value(datatype):
            print(f"KafkaDataType has no value  {datatype}")
            return False
        current_timestamp = datetime.now(timezone.utc).isoformat()
        data_dict = {"timestamp": current_timestamp,
                     "key": key,
                     "property": section,
                     "value": value,
                     "dataType": datatype.value}
        return self._kafka_client.send_message("dataflow.fct.knowledge", key, data_dict)

    def delete_keys(self, keys) -> int:
        """ Deletes all knowledge keys for the given list

              Args:
                  keys: key list
              Returns:
                  deleted documents
              """
        return self._opendata.delete_knowledge_keys(keys)

    def delete_sections(self, sections, key) -> bool:
        """ Deletes all sections from the list from the given key

               Args:
                   sections: section list
                   key: the name of the key from which the sections are to be removed

               Returns:
                   result - true if the operation was successful, false on error
               """
        return self._opendata.delete_knowledge_sections(sections, key)



from nazca4sdk.sdk import SDK
import time

sdk = SDK(False)

start_time = time.time()
data = sdk.variable_over_time('symulator', ['V1'], 15, 'DAY')
print(data)
print(type(data))
print("--- %s seconds ---" % round((time.time() - start_time), 2))
print(len(data))

start_time = time.time()
dask_data = sdk.variable_over_time('symulator', ['V1'], 15, 'DAY', dask=True)
print('----------------------')
print(dask_data)
print(type(dask_data))
print("--- %s seconds ---" % round((time.time() - start_time), 2))
print(len(dask_data))
print(dask_data.head(10))

start_time = time.time()
data_day = data = sdk.variable_over_day('symulator', ['V1'], '2023-01-15T07:00:00', '2023-02-06T07:30:00')
print('----------------------')
print(data_day)
print(type(data_day))
print("--- %s seconds ---" % round((time.time() - start_time), 2))
print(len(data_day))

start_time = time.time()
data_dask_day = sdk.variable_over_day('symulator', ['V1'], '2023-01-15T07:00:00', '2023-02-06T07:30:00', dask=True)
print('----------------------')
print(data_dask_day)
print(type(data_dask_day))
print("--- %s seconds ---" % round((time.time() - start_time), 2))
print(len(data_dask_day))
print(data_dask_day.head(10))




""" Data_mod module"""
import pandas as pd
from dask import dataframe


class Data:
    """
    Modification structure of data received from OpenData
    """

    def __init__(self, variable):
        self.__data = variable

    def to_df(self):
        """
        Returns DataFrame from received input

        Example:
            df = Data(selected_data).to_df()
        """
        if self.__is_empty():
            return []
        data_df = pd.json_normalize(self.__data)

        try:
            if 'measureTime' in data_df.columns:
                return data_df.sort_values(by='measureTime', ascending=True, ignore_index=True)
        except NameError:
            print('Error, measureTime column missing in dataframe')

    def data_to_list(self):
        """
        Method to modify data structure to list
        """

        return self.__data.values.tolist()

    def data_to_numpy(self):
        """
        Method to modify data structure to numpy array
        """

        return self.__data.to_numpy()

    def __is_empty(self):
        return len(self.__data) == 0

    def to_dask_df(self):

        if self.__is_empty():
            return []

        d1 = self.__data[0]
        ds = []

        for i in range(len(self.__data)):
            ds.append(self.__data[i])

        d = {}
        for k in d1.keys():
            d[k] = tuple(d[k] for d in ds)

        data_df = dataframe.from_dict(d, orient="columns", npartitions=1).reset_index(drop=True)

        try:
            if 'measureTime' in data_df.columns:
                data_df = data_df.sort_values(by='measureTime', ascending=True, ignore_index=True).reset_index(
                    drop=True)
                return data_df
        except NameError:
            print('Error, measureTime column missing in dataframe')

    def to_pivot_df(self):
        if self.__is_empty():
            return []
        try:
            CH_COLUMNS = ["Module", "MeasureTime", "Variable", "Value"]
            df = pd.DataFrame(self.__data, columns=CH_COLUMNS)
        except NameError:
            print('Error, name of columns are incorrect')

        df["Value"] = pd.to_numeric(df.Value)
        return df.pivot_table(index='MeasureTime', columns='Variable', values='Value')



"""Analytics module"""
from nazca4sdk.analytics.performance_indicators.indicators import Indicators
from nazca4sdk.analytics.energy.energy import Energy
from nazca4sdk.analytics.vibration.vibration_quality import VibrationQuality
from nazca4sdk.analytics.prediction.forecasting import Forecasting
from nazca4sdk.system.system_cache import SystemCache
from dependency_injector.wiring import inject


class Analytics:
    """
    Analytics module as a second layer of analytics functions to use with SDK
    """

    @inject
    def __init__(self, cache: SystemCache):
        #: :class:`nazca4sdk.analytics.performance_indicators.indicators.Indicators` Module for determining key performance indicators (such as OEE, Cp, Cpk, etc.)
        self.kpi = Indicators()
        #: :class:`nazca4sdk.analytics.energy.energy.Energy` Module for calculations related to electricity (such as energy quality)
        self.energy = Energy()
        #: :class:`nazca4sdk.analytics.vibration.vibration_quality.VibrationQuality` Module for calculations related to vibrations (such as vibration quality)
        self.vibration = VibrationQuality()
        #: :class:`nazca4sdk.analytics.prediction.forecasting.Forecasting` Module for time series forecasting
        self.forecasting = Forecasting(cache)



"""Module to calculate process indicators"""
from pydantic import ValidationError
from nazca4sdk.analytics.__brain_connection import BrainClient
from nazca4sdk.datahandling.variable_verificator import CpCpkIndicatorsParams, PpPpkIndicatorsParams


class ProcessIndicators:
    """ Class to perform process indicators calculation with cooperation with Brain """

    def __init__(self):
        self.indicators_brain = BrainClient()

    def get_cp_indicator(self, cp_input: dict):
        """
        Function to determine Cp indicator values for determined input

        Args:
            cp_input: dictionary with input parameters

        Returns:
            Cp indicator

       """

        try:
            data = dict(CpCpkIndicatorsParams(**cp_input))
            response = self.indicators_brain.get_cp_indicator(data)
            result = self.indicators_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_pp_indicator(self, pp_input: dict):
        """
        Function to determine Pp indicator values for determined input

        Args:
            pp_input: dictionary with input parameters

        Returns:
            Pp indicator

       """

        try:
            data = dict(PpPpkIndicatorsParams(**pp_input))
            response = self.indicators_brain.get_pp_indicator(data)
            result = self.indicators_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_cpk_indicator(self, cpk_input: dict):
        """
        Function to determine Cpk indicator values for determined input

        Args:
            cpk_input: dictionary with input parameters

        Returns:
            Cpk indicator

        """

        try:
            data = dict(CpCpkIndicatorsParams(**cpk_input))
            response = self.indicators_brain.get_cpk_indicator(data)
            result = self.indicators_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_ppk_indicator(self, ppk_input: dict):
        """
        Function to determine Ppk indicator values for determined input

        Args:
            ppk_input: dictionary with input parameters

        Returns:
            Ppk indicator

        """

        try:
            data = dict(PpPpkIndicatorsParams(**ppk_input))
            response = self.indicators_brain.get_ppk_indicator(data)
            result = self.indicators_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None



from nazca4sdk.sdk import SDK

sdk = SDK(False)

if ('vRmsx' and 'vRmsy' and 'vRmsz') in sdk.variables.list()['symulator666']:
    vibration = sdk.analytics.vibration.calculate_vibration_quality('G2f', 'symulator666')
else:
    vibration = sdk.analytics.vibration.calculate_vibration_quality_by_params(group="G1r", vibration=10.0)
print(f'vibration = {vibration}')

energy = sdk.analytics.energy.calculate_energy_quality('symulator')
print('--------------------------')
print("energy from module: ", energy)


energy = sdk.analytics.energy.calculate_energy_quality('brain_test')
print(energy)

vibration = sdk.analytics.vibration.calculate_vibration_quality(module="brain_test",
                                                                group="G1r")
print(vibration)


from typing import List
from pydantic import BaseModel, StrictStr


class NazcaVariable(BaseModel):
    identifier: StrictStr
    description: StrictStr
    id: int
    value: object
    valueType: StrictStr


class NazcaVariables(BaseModel):
    __root__: List[NazcaVariable]

    def variables(self):
        return self.__root__



from nazca4sdk.energy_quality import get_energy_quality
from nazca4sdk.energy_quality import EnergyInfo
import json

data = {'freq1': 50,
        "vol1": 230,
        "cos1": 1,
        "thd1": 0,
        "freq2": 50,
        "vol2": 230,
        "cos2": 1,
        "thd2": 0,
        "freq3": 50,
        "vol3": 230,
        "cos3": 1,
        "thd3": 0}

date = EnergyInfo(freq1=50, vol1=230, cos1=1, thd1=1, freq2=50, vol2=230, cos2=1, thd2=0, freq3=50, vol3=230, cos3=1,
                  thd3=0)

print(date)
EQ = get_energy_quality(
    EnergyInfo(freq1=50, vol1=230, cos1=1, thd1=1, freq2=50, vol2=230, cos2=1,
               thd2=0, freq3=50, vol3=230, cos3=1, thd3=0))

result = {'worstCaseQuality': EQ['worstCaseQuality'],
          'worstCaseQuality1': EQ['worstCaseQuality1'],
          'worstCaseQuality2': EQ['worstCaseQuality2'],
          'worstCaseQuality3': EQ['worstCaseQuality3'],
          'volQuality1': EQ['volQuality1'],
          'volQuality2': EQ['volQuality2'],
          'volQuality3': EQ['volQuality3'],
          'cosQuality1': EQ['cosQuality1'],
          'cosQuality2': EQ['cosQuality2'],
          'cosQuality3': EQ['cosQuality3'],
          'thdQuality1': EQ['thdQuality1'],
          'thdQuality1': EQ['thdQuality2'],
          'thdQuality1': EQ['thdQuality3']
          }
print(result)

# print(type(json.dumps(result)))
# time_usage = 'Time usage is equal: {}'.format(a)
# print(time_usage)
# print(b)
# print(c)


from nazca4sdk.datahandling.nazcavariables.nazca_variable import NazcaVariables, NazcaVariable
from nazca4sdk.datahandling.open_data_client import OpenDataClient
from dependency_injector.wiring import inject


class NazcaVariablesStorage:

    @inject
    def __init__(self,  opendata_client: OpenDataClient):
        self._opendata = opendata_client

    def read(self, name: str = None):
        """Read nazca variables

            Args:
                name: name of variable, if None return all variables
            Returns:
                if name == None return all nazca variables
                if name set then return nazca variable with specified name
        """

        if name is None:
            return self._read_variables()
        else:
            return self._read_variable(name)

    def _read_variables(self):
        """  Read all nazca variables

              Returns:
                  all nazca variables
        """
        response = self._opendata.read_nazca_variables()
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("Read nazca variable failure")
                return None
            variables_list = NazcaVariables.parse_raw(response.content)
            return variables_list.variables()
        print("Read nazca variables error")
        return None

    def _read_variable(self, identifier: str):
        """ Read nazca variable with identifier

                Args:
                    identifier: nazca variable identifier
                Returns:
                    nazca variable
        """

        response = self._opendata.read_nazca_variable(identifier)
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("Read nazca variable failure")
                return None
            variable = NazcaVariable.parse_raw(response.content)
            return variable
        if response.status_code == 404:
            print(f"Nazca variable {identifier} not found")
            return None
        print(f"Read nazca variable {identifier} error")
        return None



from typing import List

from pydantic import BaseModel


class BootProfileMeter(BaseModel):
    id: str
    systemName: str


class BootProfile(BaseModel):
    id: str
    name: str
    startTime: str
    recorded: bool
    isActive: bool
    meters: List[BootProfileMeter]


class BootProfiles(BaseModel):
    __root__: List[BootProfile]

    def boot_profiles(self):
        return self.__root__



"""Variable verification module"""
from datetime import datetime
from typing import Optional, List
import sys
from pydantic import BaseModel, validator

DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


class ElectricMetersDataValidator(BaseModel):
    """Class to verify the correctness of data for get_electric_meters_data function

    Attributes:
    system_name: List[str]
    variables: List[str]
    """
    system_name: List[str]
    variables: List[str]


class PgAvgDataValidator(BaseModel):
    """Class to verify the correctness of data for get_pgavg_data function

    Attributes:
    connection: str
    start_date: str
    end_date: str
    """
    connection: str
    start_date: str
    end_date: str

    @validator('start_date')
    def start_date_verify_format(cls, valid_start_date):
        """Validator to check if start_date is correct

        Args:
            valid_start_date: str

        Returns:
            valid_start_date: str
        """
        return PgAvgDataValidator.valid_datetime(valid_start_date)

    @validator('end_date')
    def end_date_verify_format(cls, valid_end_date, values):
        """Validator to check if end_date is correct

        Args:
            valid_end_date: str
            values : str
        Returns:
            valid_end_date: str
        """
        PgAvgDataValidator.valid_datetime(valid_end_date)
        if 'start_date' in values:
            start_date_str = values['start_date']
            end_date = datetime.strptime(valid_end_date, DATE_TIME_FORMAT)
            start_date = datetime.strptime(start_date_str, DATE_TIME_FORMAT)
            if end_date > start_date:
                return valid_end_date
            raise ValueError(
                f'{valid_end_date} must be after {start_date_str}')
        raise ValueError('Cannot check if end_date is after start_date')

    @staticmethod
    def valid_datetime(valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                'Bad datetime format. Format yyyy-mm-ddTHH:MM:SS')


class SsvSopOrderedPowerDataValidator(BaseModel):
    """Class to verify the correctness of data for get_ssv_sop_orderedpower function

    Attributes:
    connection: str
    """
    connection: str


class OrganizationCo2DataValidator(BaseModel):
    """Class to verify the correctness of data for get_organization_co2 function

    Attributes:
    organization_id: str
    """
    organization_id: str


class OzeConnectionInvertersDataValidator(BaseModel):
    """Class to verify the correctness of data for get_oze_connection_inverters function

    Attributes:
    connection_id: str
    """
    connection_id: str


class OzeConnectionsForMetersDataValidator(BaseModel):
    """Class to verify the correctness of data for get_oze_connections_for_meters function

    Attributes:
    meters: list
    """
    meters: list


class OzeInvertersSummaryDataValidator(BaseModel):
    """Class to verify the correctness of data for get_inverters_summary_data function

    Attributes:
    inverters: list
    from_date: str
    to_date: str
    """
    inverters: list
    from_date: str
    to_date: str


class MetersDataValidator(BaseModel):
    """Class to verify the correctness of data for get_meters_data function

    Attributes:
    meters: list
    from_date: str
    to_date: str
    """
    meters: list
    from_date: str
    to_date: str


class VariableOverDayValidator(BaseModel):
    """Class to verify the correctness of data for variable_over_day function

    Attributes:
    module_name: str
    variable_names: list
    """
    module_name: str
    variable_names: list


class VariableIntervalInfo(VariableOverDayValidator):
    """Class to verify the correctness of data for variable over time range
    when start and end date is set

    Attributes:
        module parameters: dict
    """
    start_date: str
    end_date: str

    @validator('start_date')
    def start_date_verify_format(cls, valid_start_date):
        """Validator to check if time amount is correct

        Args:
            valid_start_date: str

        Returns:
            valid_start_date: str
        """
        return VariableIntervalInfo.valid_datetime(valid_start_date)

    @validator('end_date')
    def end_date_verify_format(cls, valid_end_date, values):
        """Validator to check if time amount is correct

        Args:
            valid_end_date: str
            values : str
        Returns:
            valid_end_date: str
        """
        VariableIntervalInfo.valid_datetime(valid_end_date)
        if 'start_date' in values:
            start_date_str = values['start_date']
            end_date = datetime.strptime(valid_end_date, DATE_TIME_FORMAT)
            start_date = datetime.strptime(start_date_str, DATE_TIME_FORMAT)
            if end_date > start_date:
                return valid_end_date
            raise ValueError(
                f'{valid_end_date} must be after {start_date_str}')
        raise ValueError('Cannot check if end_date is after start_date')

    @staticmethod
    def valid_datetime(valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                'Bad datetime format. Format yyyy-mm-ddTHH:MM:SS')


class VariableIntervalSubtractionInfo(BaseModel):
    """Class to verify the correctness of data for variable over time range
     when time unit and time amount is set

    Attributes:
        time parameters: dict
    """
    time_unit: str
    time_amount: int

    @validator('time_amount')
    def time_amount_validator(cls, valid_time_amount):
        """
        Validator to check if time amount is correct

        Args:
            valid_time_amount: int

        Returns:
            valid_time_amount: int
        """
        if 0 >= valid_time_amount > sys.maxsize:
            raise ValueError('time_amount has to be greater than 0')
        return valid_time_amount

    @validator('time_unit')
    def time_unit_validator(cls, valid_time_unit):
        """Validator to check if time unit is correct

        Args:
            valid_time_unit: str

        Returns:
            valid_time_unit: str
        """
        possibilities = ['SECOND', 'MINUTE',
                         'HOUR', 'DAY', "MONTH", 'WEEK', 'YEAR']
        if valid_time_unit not in possibilities:
            raise ValueError(
                'Wrong time aggregator, try: SECOND, MINUTE, HOUR, DAY, WEEK, YEAR')
        return valid_time_unit


class OeeSimpleParams(BaseModel):
    """Class to verify the correctness of data for Simple OEE function

    Attributes:
    availability: float
    performance: float
    quality: float
    """
    availability: float
    performance: float
    quality: float

    @validator('availability')
    def availability_validator(cls, value):
        """ Availability parameter validator """
        if value < 0:
            raise ValueError('Availability has to be higher than')
        return float(value)

    @validator('performance')
    def performance_validator(cls, value):
        """ Performance parameter validator """
        if value < 0:
            raise ValueError('Performance has to be higher than')
        return float(value)

    @validator('quality')
    def quality_validator(cls, value):
        """Quality parameter validator """
        if value < 0:
            raise ValueError('Quality has to be higher than')
        return float(value)


class OeeComplexParams(BaseModel):
    """Class to verify the correctness of data for variable OEE

    Attributes:
    A : (float) Total available time,
    B : (float) Run time,
    C : (float) Production capacity,
    D : (float) Actual production,
    E : (float) Production output (same as actual production),
    F : (float) Actual good products (i.e. product output minus scraps)
    """

    A: float
    B: float
    C: float
    D: float
    E: float
    F: float


class AvailabilityValidator(BaseModel):
    """Class to verify the correctness of data for availability

    Attributes:
    run_time: float
    total_time: float
    """
    run_time: float
    total_time: float


class PerformanceValidator(BaseModel):
    """Class to verify the correctness of data for performance

    Attributes:
    actual_production: float
    production_capacity: float
    """
    actual_production: float
    production_capacity: float


class QualityValidator(BaseModel):
    """Class to verify the correctness of data for quality

    Attributes:
    actual_products: float
    production_output: float
    """
    actual_products: float
    production_output: float


class CpCpkIndicatorsParams(BaseModel):
    """
    Class to verify the correctness of data for cp, cpk indicators

    Args:
    module: (str)
    variable: (str)
    start_date: (str)
    end_date: (str)
    lsl: (float)
    usl: (float)
    period: (int)
    subgroups: (int)
    estimation_type: (str)
    """
    module: str
    variable: str
    start_date: str
    end_date: str
    lsl: Optional[float]
    usl: Optional[float]
    period: int
    subgroups: int
    estimation_type: str

    @validator('estimation_type')
    def estimation_type_validate(cls, estimation_type):
        """ estimation_type validator """
        if estimation_type != 'time' and estimation_type != 'samples':
            raise ValueError('estimation_type has to be time or samples')

        return estimation_type

    @validator('subgroups')
    def subgroups_validate(cls, subgroups):
        """ subgroups validator """
        if subgroups <= 0:
            raise ValueError('Value has to be greater than 0')

        return float(subgroups)

    @validator('period')
    def period_validate(cls, period):
        """ period validator """
        if period <= 0:
            raise ValueError('Value has to be greater than 0')

        return float(period)

    @validator('usl', pre=True, always=True)
    def limits_validate(cls, usl, values):
        """ Cpk/ppk limits validator """
        if not values.get('lsl') and not usl:
            raise ValueError('Either lsl or usl is required')

        if not usl:
            return

        if 'lsl' in values:
            lsl_value = values['lsl']

            if lsl_value is None:
                return usl
            if usl > lsl_value:
                return usl
            raise ValueError('Value has to be grater than lsl')

    @validator('start_date')
    def start_date_verify_format(cls, valid_start_date):
        """Validator to check if time amount is correct

        Args:
            valid_start_date: str

        Returns:
            valid_start_date: str
        """
        return VariableIntervalInfo.valid_datetime(valid_start_date)

    @validator('end_date')
    def end_date_verify_format(cls, valid_end_date, values):
        """Validator to check if time amount is correct

        Args:
            valid_end_date: str
            values : str
        Returns:
            valid_end_date: str
        """
        VariableIntervalInfo.valid_datetime(valid_end_date)
        if 'start_date' in values:
            start_date_str = values['start_date']
            end_date = datetime.strptime(valid_end_date, DATE_TIME_FORMAT)
            start_date = datetime.strptime(start_date_str, DATE_TIME_FORMAT)
            if end_date > start_date:
                return valid_end_date
            raise ValueError(
                f'{valid_end_date} must be after {start_date_str}')
        raise ValueError('Cannot check if end_date is after start_date')


class PpPpkIndicatorsParams(BaseModel):
    """
    Class to verify the correctness of data for pp, ppk indicators

    Attributes:
    module: (str)
    variable: (str)
    start_date: (str)
    end_date: (str)
    lsl : (float)
    usl : (float)


    """
    module: str
    variable: str
    start_date: str
    end_date: str
    lsl: Optional[float]
    usl: Optional[float]

    @validator('usl')
    def limits_validate(cls, usl, values):
        """ Cpk/ppk limits validator """
        if not values.get('lsl') and not usl:
            raise ValueError('Either lsl or usl is required')

        if not usl:
            return

        if 'lsl' in values:
            lsl_value = values['lsl']

            if lsl_value is None:
                return usl
            if usl > lsl_value:
                return usl
            raise ValueError('Value has to be grater than lsl')

    @validator('start_date')
    def start_date_verify_format(cls, valid_start_date):
        """Validator to check if time amount is correct

        Args:
            valid_start_date: str

        Returns:
            valid_start_date: str
        """
        return VariableIntervalInfo.valid_datetime(valid_start_date)

    @validator('end_date')
    def end_date_verify_format(cls, valid_end_date, values):
        """Validator to check if time amount is correct

        Args:
            valid_end_date: str
            values : str
        Returns:
            valid_end_date: str
        """
        VariableIntervalInfo.valid_datetime(valid_end_date)
        if 'start_date' in values:
            start_date_str = values['start_date']
            end_date = datetime.strptime(valid_end_date, DATE_TIME_FORMAT)
            start_date = datetime.strptime(start_date_str, DATE_TIME_FORMAT)
            if end_date > start_date:
                return valid_end_date
            raise ValueError(
                f'{valid_end_date} must be after {start_date_str}')
        raise ValueError('Cannot check if end_date is after start_date')


class VariableType(BaseModel):
    """Class to verify the type of variable names

    Attributes:
        variables_grouped: variable type
    """
    variables_grouped: list

    @validator('variables_grouped')
    def variable_verify_type(cls, variables_grouped):
        """Validator to check the value type of input variable names

        Args:
            variables_grouped: str

        Returns:
            variables_grouped: str
        """
        if variables_grouped[0][0] != 'float' and variables_grouped[0][0] != 'int':
            raise ValueError(
                'Values from input variable has to be float or int')

        return variables_grouped[0][1][0]


class ForecastForPrediction(BaseModel):
    """Class to verify the forecast time amount ant unit to make future dataframe to for prediction and verify
    prediction tool

    Attributes:
        forecast_time_amount: int
        forecast_time_unit: str
        prediction_tool: str: "prophet" or "nixtla"
    """

    forecast_time_amount: int
    forecast_time_unit: str
    prediction_tool: str

    @validator('prediction_tool')
    def prediction_tool_verify_name(cls, prediction_tool):
        """Validator to check the prediction_tool

        Args:
            prediction_tool: str

        Returns:
            prediction_tool: str
        """

        tools = ['prophet', 'nixtla']

        if prediction_tool not in tools:
            raise ValueError('prediction tool has to be "prophet" or "nixtla"')

        return prediction_tool

    @validator('forecast_time_amount')
    def forecast_time_amount_verify(cls, forecast_time_amount):
        """Validator to check the forecast_time_unit

        Args:
            forecast_time_amount: str

        Returns:
            forecast_time_amount: str
        """

        if 0 >= forecast_time_amount > sys.maxsize:
            raise ValueError('forecast_time_amount has to be greater than 0')
        return forecast_time_amount

    @validator('forecast_time_unit')
    def forecast_time_unit_verify(cls, forecast_time_unit):
        """Validator to check the forecast_time_unit

        Args:
            forecast_time_unit: str

        Returns:
            forecast_time_unit: str
        """

        units = ['Y', 'M', 'D', 'H', 'MIN', 'S']

        if forecast_time_unit.upper() not in units:
            raise ValueError("forecast time unit has to be 'Y' or 'y' (year), 'M' or 'm' (month), 'D' or 'd' (day), "
                             "'H' or 'h' (hour), 'MIN' or 'min' (minute), 'S' or 's' (second)")

        return forecast_time_unit



import functools
import textwrap


class Query:
    keywords = [
        'SELECT',
        'FROM',
        'WHERE',
        'GROUP BY',
        'ORDER BY',
        'LIMIT',
        'OFFSET'
    ]

    def __init__(self):
        self.data = {k: [] for k in self.keywords}

    def add(self, keyword, arg: str):
        expression = self.data[keyword]
        expression.append(Query._clean_up(arg))
        return self

    def statement(self) -> str:
        return str(self)

    def __getattr__(self, name):
        if not name.isupper():
            return getattr(super(), name)
        return functools.partial(self.add, name.replace('_', ' '))

    def __str__(self):
        return ''.join(self._lines())

    def _lines(self):
        for keyword, expression in self.data.items():
            if not expression:
                continue

            yield f'{keyword} {expression[0]} '

    @staticmethod
    def _clean_up(thing: str) -> str:
        delete = thing.split(';')[0]
        res = delete.split()
        return textwrap.dedent(" ".join(res))



import json
from kafka import KafkaProducer, errors


class KafkaSender:

    def __init__(self, host: str, port: str):
        self._producer = None
        self._kafka_host = host
        self._kafka_port = int(port)
        # self._producer = KafkaProducer(bootstrap_servers='10.217.1.201:10092')

    def send_message(self, topic: str, key: str, data: object):
        try:
            producer = KafkaProducer(bootstrap_servers=f'{self._kafka_host}:{self._kafka_port}')
            d = json.dumps(data)

            result = producer.send(topic=topic, key=bytes(
                key, 'utf-8'), value=str.encode(d))
            res = result.get(timeout=15)
            print(res)
            producer.flush()
            producer.close(5)
            return True
        except errors.KafkaError as e:
            print(e)
        return False



from nazca4sdk.datahandling.variable_verificator import VariableIntervalInfo
from pydantic import validator

MAX_PAGE_SIZE = 10000


class VariableHistoricalInfo(VariableIntervalInfo):
    """
    Class to verify the correctness of data for sdk.read_historical_variable range
    when start and end date is set

    Attributes:
        module parameters: dict
    """
    page_size: int = MAX_PAGE_SIZE
    page_number: int = 1

    @validator('page_size')
    def page_size_validator(cls, valid_page_size):
        if valid_page_size > MAX_PAGE_SIZE:
            raise ValueError(f"Page size can be max {MAX_PAGE_SIZE}")
        return valid_page_size



from pydantic import BaseModel, StrictStr, validator
from typing import List


class VariableStats(BaseModel):
    module: str
    variable: str
    min: float
    max: float
    avg: float
    lastValue: float
    variance: float
    std: float


class VariablesStats(BaseModel):
    __root__: List[VariableStats]

    def stats(self):
        return self.__root__



from nazca4sdk.datahandling.hotstorage.model.user_variable import UserVariable, UserVariableInfo
from nazca4sdk.datahandling.hotstorage.model.user_variable import UserVariableDataType
from nazca4sdk.datahandling.hotstorage.user_variable_value import UserVariableValue
from nazca4sdk.datahandling.hotstorage.user_variable_value import user_variable_data_types
from nazca4sdk.datahandling.hotstorage.user_variables_info_value import UserVariableInfoValue

table_name_dict = {
    "bool": "bool",
    "int": "int",
    "double": "float",
    "text": "string",
    "datetime": "datetime"}

userVariableDataTypeNames = {
    UserVariableDataType.BOOL: "bool",
    UserVariableDataType.INT: "int",
    UserVariableDataType.DOUBLE: "double",
    UserVariableDataType.TEXT: "text",
    UserVariableDataType.DATETIME: "datetime"}

userVariableTableNames = {
    UserVariableDataType.BOOL: "bool",
    UserVariableDataType.INT: "int",
    UserVariableDataType.DOUBLE: "float",
    UserVariableDataType.TEXT: "string",
    UserVariableDataType.DATETIME: "datetime"
}


def get_variable_table_name(user_variable_data_type: UserVariableDataType):
    if user_variable_data_type in userVariableTableNames:
        return userVariableTableNames[user_variable_data_type]
    return None


def transform(user_variables: [UserVariable]):
    variables = []
    for user_variable in user_variables:
        try:
            user_variable_data_type = userVariableDataTypeNames[user_variable.type]
            new_var = UserVariableValue(name=user_variable.name,
                                        type=user_variable_data_type,
                                        dateTime=user_variable.datetime,
                                        value=user_variable.value)
            variables.append(new_var)
        except KeyError as e:
            print(f"UserVariableInfo(name={user_variable.name}) type error. Type {e} not supported")
    return variables


def get_table_name(data_type: str):
    try:
        table = table_name_dict[data_type]
        return table
    except KeyError:
        return ""


def transform_info(user_variables: [UserVariableInfo]):
    variables = []
    for user_variable_info in user_variables:
        try:
            user_variable_data_type = userVariableDataTypeNames[user_variable_info.type]
            new_var = UserVariableInfoValue(name=user_variable_info.name,
                                            type=user_variable_data_type)
            variables.append(new_var)
        except KeyError as e:
            print(f"UserVariableInfo(name={user_variable_info.name}) type error. Type {e} not supported")
    return variables


def create_variables_frame(user_variables: [UserVariableValue]):
    variables_frame = []
    for variable_type in user_variable_data_types:
        data_type_variables = filter(lambda x: x.type == variable_type, user_variables)
        variables_list = list(data_type_variables)
        if len(variables_list) > 0:
            variables_data = []
            table_name = get_table_name(variable_type)
            if not table_name:
                return ""

            variables_info = {"tableName": f"user_data_{table_name}", "data": variables_data}
            for el in variables_list:
                e = [el.dateTime.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                     el.dateTime.strftime("%Y-%m-%dT%H:%M:%S"),
                     el.name]
                if variable_type == "bool":
                    e.append(str(int(el.value)))
                else:
                    e.append(str(el.value))

                variables_data.append(e)
            variables_frame.append(variables_info)
    return variables_frame



from nazca4sdk import SDK

sdk = SDK(False)

test = sdk.npoe.check_boot_profiles_to_record()
connections = sdk.npoe.get_connections()

# # get electric meters data
# electric_meters1 = sdk.npoe.get_electric_meters_data(system_name=["apa260"], time_amount=10, time_unit="HOUR")
# electric_meters2 = sdk.npoe.get_electric_meters_data(system_name=["apa260"],
#                                                      start_date='2023-08-07T00:00:00',
#                                                      end_date='2023-08-08T00:00:00')
# electric_meters3 = sdk.npoe.get_electric_meters_data(system_name=["apa260"],
#                                                      time_amount=10,
#                                                      time_unit="HOUR",
#                                                      variables=["U1", "U2"])
# print("electric_meters1: ", electric_meters1.head(2))
# print("electric_meters2: ", electric_meters2.head(2))
# print("electric_meters3: ", electric_meters3.head(2))

pg_max = sdk.npoe.calculate_pt_max_window()


pgavg_data = sdk.npoe.get_pgavg_data(connection="p_apa260",
                                     start_date='2023-08-07T00:00:00.345Z',
                                     end_date='2023-08-31T00:00:00.000Z')
print("pgavg: \n\n", pgavg_data.head(2))

sop_ssv_op_data = sdk.npoe.get_ssv_sop_orderedpower("p_apa260")
print("sop_ssv_op_data: \n\n", sop_ssv_op_data.head())

occ = sdk.analytics.energy.calculate_optimal_contracted_capacity(start_date='2023-08-01T00:00:00.398Z',
                                                                 end_date='2023-08-31T00:00:00.000Z',
                                                                 connection='p_apa260', simulated_ordered_power=20,
                                                                 ordered=False, simulated=True)
print(occ)


from nazca4sdk.datahandling.knowledge.knowledge_data_type import KnowledgeDataType
from nazca4sdk.sdk import SDK
from nazca4sdk import FormatChart, FormatText
import requests
import datetime


sdk = SDK()
# odczyt dla danego klucza

# delete knowledge
sdk.knowledge.write_key_values("pomidor1", "sekcja pomidor1", "test", KnowledgeDataType.TEXT)
sdk.knowledge.write_key_values("pomidor1", "sekcja pomidor2", "test", KnowledgeDataType.TEXT)
sdk.knowledge.write_key_values("pomidor1", "sekcja pomidor3", "test", KnowledgeDataType.TEXT)

result = sdk.knowledge.delete_sections(["sekcja pomidor1", "sekcja pomidor2"], "pomidor1")

result3 = sdk.knowledge.write_key_values("pomidor1", "sekcja pomidor", "test", KnowledgeDataType.TEXT)
result2 = sdk.knowledge.write_key_values("pomidor2", "sekcja pomidor", "test", KnowledgeDataType.TEXT)
deleted_documents = sdk.knowledge.delete_keys(["pomidor1", "pomidor2"])

values = sdk.knowledge.read_keys("test")
keys = sdk.knowledge.read_keys("test", "2023-01-01T00:00:00", "2023-01-11T00:00:00", "0")
keys2 = sdk.knowledge.read_keys("test", "2023-01-01T00:00:00")
keys3 = sdk.knowledge.read_keys("test")
keys4 = sdk.knowledge.read_keys(size=4)
# print(values)

# result = sdk.write_knowledge("blob", "pliczek", "/2022-07-07_14-31.png", KnowledgeDataType.BLOB)
# result = sdk.write_knowledge("gruby", "sekcjaGruby", "test", KnowledgeDataType.TEXT)
# print(result)



import re
from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from typing import Optional


class UserVariableDataType(Enum):
    BOOL = 0
    INT = 1
    DOUBLE = 2
    TEXT = 3
    DATETIME = 4


@dataclass()
class UserVariable:
    type: UserVariableDataType
    name: str
    value: object
    datetime: Optional[datetime] = None

    def __post_init__(self):
        if self.datetime is None:
            self.datetime = datetime.now()
        if self.type == UserVariableDataType.DATETIME and isinstance(self.value, str):
            if re.search("[+-][0-9][0-9]:[0-9][0-9]", self.value):
                raise ValueError("DataTime value can't contain time offset")


@dataclass()
class UserVariableInfo:
    type: UserVariableDataType
    name: str



"""Module to calculate OEE"""
from pydantic import ValidationError
from nazca4sdk.analytics.__brain_connection import BrainClient
from nazca4sdk.datahandling.variable_verificator import OeeSimpleParams, OeeComplexParams, \
    AvailabilityValidator, PerformanceValidator, QualityValidator


class Oee:
    """ Class to perform Oee calculation with cooperation with Brain """

    def __init__(self):
        self.oee_brain = BrainClient()

    def calculate_oee_simple(self, simple_input: dict):
        """
        Function to determine energy quality values for determined input

        Args:
        simple_input -> dictionary with simple oee parameters:
            availability: (float) proces availability,
            performance: (float) proces performance,
            quality: (float) proces quality

        Returns:
            oee parameter: float
       """

        try:
            data = dict(OeeSimpleParams(**simple_input))
            response = self.oee_brain.get_oee_easy(data)
            result = self.oee_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def calculate_oee_complete(self, complex_input: dict):
        """
        Function to determine energy quality values for determined input

            Args:
            complex_input -> dictionary with complex oee parameters::
                availability,
                performance,
                quality

        Returns:
            float: oee parameter
        """

        try:
            data = dict(OeeComplexParams(**complex_input))
            response = self.oee_brain.get_oee_full(data)
            result = self.oee_brain.parse_response(response)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    @staticmethod
    def calculate_availability(availability_input: dict):
        """
        Availability:
        Takes into account availability/time loss, which includes all events
        related to unplanned stops
        (e.g. equipment failures, material shortages) and planned stops
        (e.g. changeover times).
        Availability measures the proportion of time a machine or cell runs from
        the total theoretical available time.

        Calculated as:

        Availability = Run time/Total available time
        Args:
            ::availability_input -> dictionary with availability parameters::
            run_time,
            total_time,

        Returns:
            Availability value: float
        """

        try:
            data = AvailabilityValidator(**availability_input)
            result = data.run_time / data.total_time
            return result
        except ValidationError as error:
            print(error.json())
            return None

    @staticmethod
    def calculate_performance(performance_input: dict):
        """Performance:Takes into account performance/speed loss,
        which includes all the factors
        (e.g. slow cycles, small steps)
        that prevent the machine or cell to operate at maximum/optimal speed.
        It measures the proportion of produced units from the total number of
        possible produced units in a given run.

        Calculated as:

        Performance = Actual production/Production capacity

        Args:
        :performance_input -> dictionary with performance parameters:
            actual_production, Actual production
            production_capacity, Production capacity

        Returns:
            Performance value: float
        """

        try:
            data = PerformanceValidator(**performance_input)
            result = data.actual_production / data.production_capacity
            return result
        except ValidationError as error:
            print(error.json())
            return None

    @staticmethod
    def calculate_quality(quality_input: dict):
        """
        Quality:
        Takes into account quality loss, which includes all the factors
        (e.g. reworks, scraps, defects)
        that lead to defective units that do not meet the customer’s quality
        standards and specifications.
        Quality measures the proportion of non-defective units
        compared to the total units produced.

        Calculated as:

        Quality = Actual good products/Product output
        Args:
            :quality_input -> dictionary with performance parameters::
            actual_products, Actual good products
            production_output, Production output

        Returns:
            Quality valueL: float
        """

        try:
            data = QualityValidator(**quality_input)
            result = data.actual_products / data.production_output
            return result
        except ValidationError as error:
            print(error.json())
            return None



"""OpenData Client module"""
import datetime
import io
import json
from typing import List, Optional

import requests
from requests.exceptions import ConnectionError

from nazca4sdk.datahandling.cache.key_value import KeyValue
from nazca4sdk.datahandling.data_mod import Data
from nazca4sdk.datahandling.hotstorage.helper import get_variable_table_name
from nazca4sdk.datahandling.hotstorage.user_variable_value import UserVariableValue, UserVariablesValues
from nazca4sdk.datahandling.hotstorage.user_variables_info_value import UserVariablesInfoValues
from nazca4sdk.datahandling.hotstorage.user_variables_stats_info import UserVariablesStatsInfo
from nazca4sdk.datahandling.hotstorage.variables_stats_info import VariableStatsInfo
from nazca4sdk.datahandling.hotstorage.helper import get_variable_table_name
from nazca4sdk.datahandling.npoe.boot_profiles import BootProfiles, BootProfile
from nazca4sdk.datahandling.npoe.connection import Connections
from dependency_injector.wiring import inject

JSON_HEADERS = {'Content-Type': 'application/json',
                'Accept': 'application/json'}


class DateTimeAwareEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


class OpenDataClient:
    """
    Get data from OpenData

    """
    @inject
    def __init__(self, url: str,  https: str):
        """
        Initialize OpenData url to receive system configuration
        """
        if https == "True":
            self.base_url = f'https://{url}:5001'
        else:
            self.base_url = f'http://{url}:80'

    @staticmethod
    def parse_response(response):
        """
        Parsing method

        """
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("OpenData response failure")
                return None
            return Data(json_response)

        print("OpenData response error")
        return None

    def get_paged_data(self, url: str, params: list, page_size=10000):
        """
        Get request to url with params and return result array of page

        Args:
            url: url to endpoint
            params: params to get request
            page_size: size of page

        Returns:
            array of  values : Paged values
        Example:
            get_page_data(url='/api/hotstorage/ReadHistoricalVariable', params=[('start_date','2022-12-16T00:00:00')],
            page_size=100)
        """

        api_url = f'{self.base_url}{url}'
        page_number = 1
        result = []
        session = requests.Session()
        headers = JSON_HEADERS
        try:
            params.append(('pageSize', page_size))
            p = params.copy()
            p.append(('pageNumber', page_number))
            response = session.get(
                api_url,
                params=params,
                headers=headers, verify=False)
            if response.status_code == 200:
                result = response.json()
            yield result
            result_page_size = len(result)
            page_number = page_number + 1
            while result_page_size == page_size:
                pa = params.copy()
                pa.append(('pageNumber', page_number))
                result = session.get(api_url, params=pa, headers=headers, verify=False).json()
                result_page_size = len(result)
                page_number = page_number + 1
                yield result
        except AttributeError:
            print('Can not mix time descriptions, use start_date end_date'
                  ' or time amount with time unit')
            return None

    def request_params(self, module_name, grouped_variables, **time):
        """
        Creates request for OpenData

        Args:
            module_name: module_name
            grouped_variables: list of variable names
            time: **time - Time definition:
                Possible pairs: start_date end_date or time_amount time_unit
        Returns:
            request for Open Data
        """
        headers = JSON_HEADERS
        try:
            if time.get("start_date") and time.get("end_date") is not None:
                request_params = [
                    ('module', module_name),
                    ('startdate', time['start_date']),
                    ('enddate', time['end_date'])
                ]

                for variable_type in grouped_variables:
                    for variable_name in variable_type[1]:
                        request_params.append((f'groupedVariables[{variable_type[0]}]',
                                               variable_name))

                response = requests.get(
                    f'{self.base_url}/api/HotStorage/VariableOverTime',
                    params=request_params,
                    headers=headers, verify=False)
                return response

        except AttributeError:
            print('Can not mix time descriptions, use start_date end_date'
                  ' or time amount with time unit')
            return None

    def save_variables(self, user_variables: [UserVariableValue]):
        if not isinstance(user_variables, list):
            raise ValueError("user variable is not a list")
        if not all(isinstance(x, UserVariableValue) for x in user_variables):
            raise ValueError("All object in user_variable list should by type UserVariable")
        headers = JSON_HEADERS
        try:
            variables = UserVariablesValues(__root__=user_variables)
            request_params = variables.json()
            response = requests.post(
                f'{self.base_url}/api/HotStorage/saveUserVariables',
                data=request_params,
                headers=headers, verify=False)
            return response

        except AttributeError:
            print('Can not mix time descriptions, use start_date end_date'
                  ' or time amount with time unit')
            return None

    def read_variables(self, user_variables_info: UserVariablesInfoValues):
        headers = JSON_HEADERS
        try:
            request_params = [
                ('startdate', user_variables_info.startDate),
                ('enddate', user_variables_info.endDate)
            ]

            for variable_type in user_variables_info.variables:
                for variable_name in user_variables_info.variables[variable_type]:
                    request_params.append((f'variables[{variable_type}]', variable_name))

            response = requests.get(
                f'{self.base_url}/api/HotStorage/userVariablesOverTime',
                params=request_params,
                headers=headers, verify=False)
            return response

        except AttributeError:
            print('Can not mix time descriptions, use start_date end_date'
                  ' or time amount with time unit')
            return None

    def download_file(self, path: str, local_path: str):
        """download file from file storage

        Args:
            path: path to file
            local_path: path to save the file

        Returns:
            binary of  file

        """
        headers = JSON_HEADERS
        try:
            request_params = [('path', path)]
            response = requests.get(
                f'{self.base_url}/api/FileRepository/file',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return False
        except AttributeError:
            print(f'error while downloading file {path} from file repository')
            return False
        if response.status_code == 200:
            f = open(local_path, "wb")
            f.write(bytes(response.content))
            f.close()
            return True
        print(f"Error downloading file {path}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return False

    def send_file(self, path: str, file: io.BytesIO):
        r"""send file to file storage

        Args:
            file: binary file to send
            path: path to file on file storage

        Returns:
            True - file saved, False - file save error

        """

        try:
            files = {'file': ("File.txt", file, "application/octet-stream")}
            request_params = [('path', path)]
            response = requests.post(
                f'{self.base_url}/api/FileRepository/file',
                params=request_params,
                files=files,
                verify=False)
        except ConnectionError as e:
            print(f'Failed to establish connection to OpenData {e}')
            return None
        except AttributeError:
            print(f'error while sending file {path} to file repository')
            return False
        if response.status_code == 200 or response.status_code == 204:
            print(f"file {path} saved.")
            return True
        print(f"Error sending file {path}, status code = {response.status_code}")
        return False

    def read_cache_keys(self, keys):
        headers = {'Accept': 'application/json'}
        try:

            request_params = {}
            for index in range(0, len(keys)):
                request_params[f"keys[{index}]"] = keys[index]

            response = requests.get(
                f'{self.base_url}/api/Cache',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while get values for keys - {keys}')
            return None
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("OpenData response failure")
                return None
            return json_response
        print(f"error while get values for keys - {keys}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def write_cache_keys(self, key_values: KeyValue):
        """write key value to cache
        Args:
            key_values: dict key, value
        Returns:
              CacheEntry json if value saved or None if error
        """
        headers = JSON_HEADERS
        try:

            request_params = {"key": key_values.key,
                              "value": key_values.value}
            dump = json.dumps(request_params, cls=DateTimeAwareEncoder)
            response = requests.post(
                f'{self.base_url}/api/Cache/writeKeyValue',
                data=dump,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while write values - {key_values}')
            return None
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("OpenData response failure")
                return None
            return json_response
        print(f"error while write cache values - {key_values}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def read_knowledge_keys(self, name=None, ts_min=None, ts_max=None, size=0):
        headers = {'Accept': 'application/json'}
        try:
            request_params = {
                "key": name,
                "ts_min": ts_min,
                "ts_max": ts_max,
                "size": size
            }
            response = requests.get(
                f'{self.base_url}/api/knowledge/knowledgeKeys',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print('Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while get knowledge keys')
            return None
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("Get knowledge  failure")
                return None
            return json_response
        print(f"error while get knowledge keys, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def read_knowledge_values(self, key):
        headers = {'Accept': 'application/json'}
        try:

            request_params = {
                "key": key
            }

            response = requests.get(
                f'{self.base_url}/api/knowledge',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while get knowledge for key - {key}')
            return None
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("Get knowledge  failure")
                return None
            return json_response
        print(f"error while get knowledge for key - {key}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def delete_knowledge_keys(self, keys):
        headers = {'Accept': 'application/json'}
        try:
            request_params = {}
            for index in range(0, len(keys)):
                request_params[f"keys[{index}]"] = keys[index]

            response = requests.delete(
                f'{self.base_url}/api/knowledge/keys',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while get values for keys - {keys}')
            return None
        if response.status_code == 200:
            return int(response.content)
        print(f"error while get values for keys - {keys}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def delete_knowledge_sections(self, sections, key):
        headers = {'Accept': 'application/json'}
        try:
            if type(sections) != list:
                print('The ''sections'' argument is not a list')
                return False
            request_params = {}
            for index in range(0, len(sections)):
                request_params[f"sections[{index}]"] = sections[index]
            request_params["key"] = key

            response = requests.delete(
                f'{self.base_url}/api/knowledge/sections',
                params=request_params,
                headers=headers,
                verify=False)
        except ConnectionError:
            print(' Failed to establish connection to OpenData')
            return None
        except AttributeError:
            print(f'error while get values for keys - {key}')
            return None
        if response.status_code == 200:
            return response.content
        print(f"error while get values for keys - {key}, response.status_code={response.status_code}, "
              f"response.content={response.content}")
        return None

    def read_nazca_variables(self):
        headers = {'Accept': 'application/json'}
        response = requests.get(
            f'{self.base_url}/api/NazcaVariables', headers=headers, verify=False)
        return response

    def read_nazca_variable(self, identifier: str):
        headers = {'Accept': 'application/json'}
        response = requests.get(
            f'{self.base_url}/api/NazcaVariables/{identifier}', headers=headers, verify=False)
        return response

    def read_variables_stats(self, variables_stats_info: VariableStatsInfo):
        headers = JSON_HEADERS
        try:
            request_params = [
                ('startdate', variables_stats_info.startDate),
                ('enddate', variables_stats_info.endDate),
                ('module', variables_stats_info.module)
            ]

            for variable_type in variables_stats_info.variables:
                for variable_name in variables_stats_info.variables[variable_type]:
                    request_params.append((f'variables[{variable_type}]', variable_name))

            response = requests.get(
                f'{self.base_url}/api/HotStorage/variablesStats',
                params=request_params,
                headers=headers, verify=False)
            return response

        except AttributeError:
            print('Attribute error')
            return None

    def read_user_variables_stats(self, info: UserVariablesStatsInfo):
        headers = JSON_HEADERS
        try:
            request_params = [
                ('startdate', info.startDate),
                ('enddate', info.endDate)
            ]

            for variable in info.variables:
                var_type = get_variable_table_name(variable.type)
                if var_type is None:
                    print(f"Variable type {variable.type} not recognized")
                request_params.append((f'variables[{var_type}]', variable.name))

            response = requests.get(
                f'{self.base_url}/api/HotStorage/uservariablesStats',
                params=request_params,
                headers=headers, verify=False)
            return response

        except AttributeError:
            print('Attribute error')
            return None

    def get_boot_profiles(self) -> Optional[List[BootProfile]]:
        headers = JSON_HEADERS
        try:
            response = requests.get(
                f'{self.base_url}/api/Config/npoe/bootprofiles',
                headers=headers, verify=False)
            return BootProfiles.parse_raw(response.content).boot_profiles()

        except AttributeError:
            print('Attribute error')
            return None

    def set_boot_profiles_as_recorded(self, profile_ids: List[str]):
        headers = JSON_HEADERS
        try:
            requests.post(
                f'{self.base_url}/api/Config/npoe/bootprofiles/recorded',
                headers=headers, verify=False, json={"guids": profile_ids})

        except Exception:
            print('Attribute error')
            return None

    def get_connections(self):
        headers = JSON_HEADERS
        try:
            response = requests.get(
                f'{self.base_url}/api/Config/connections',
                headers=headers, verify=False)
            return Connections.parse_raw(response.content).connections()

        except AttributeError:
            print('Attribute error')
            return None



from dataclasses import dataclass
from nazca4sdk.datahandling.hotstorage.model.user_variable import UserVariableDataType


@dataclass()
class UserVariableInfo:
    type: UserVariableDataType
    name: str



""" Module to communicate with Brain"""
import json

import requests


class BrainClient:
    """ Get data from Nazca4.0 Brain
    """

    def __init__(self):
        """Initialize Brain connection info"""
        self.__brain_url = 'http://brain:8000'

    def get_optimal_contracted_capacity(self, start_date: str, end_date: str, connection: str,
                                        simulated_ordered_power: int,
                                        ordered: bool, simulated: bool):
        """ Posting data to receive optimal contracted capacity calculations from Brain"""
        return requests.post(f'{self.__brain_url}/optimal_contracted_capcity/', json={
                                                                    "start_date": start_date,
                                                                    "end_date": end_date,
                                                                    "connection": connection,
                                                                    "simulated_ordered_power": simulated_ordered_power,
                                                                    "ordered": ordered,
                                                                    "simulated": simulated})

    def get_energy_quality(self, module: str, npoe: bool):
        """ Posting data to receive energy quality calculations from Brain"""
        if npoe:
            return requests.post(f'{self.__brain_url}/energy/npoe', json={"meter": module})
        return requests.post(f'{self.__brain_url}/energy/nazca', json={"module_name": module})

    def get_energy_quality_by_params(self, params: dict):
        """ Posting data to receive energy quality calculations from Brain"""
        return requests.post(f'{self.__brain_url}/energy_by_params/', json=params)

    def get_vibration_quality(self, group: str, module: str):
        """ Posting data to receive vibration quality calculations from Brain"""
        return requests.post(f'{self.__brain_url}/vibration/', json={"module_name": module, "group": group})

    def get_vibration_quality_by_params(self, group: str, vibration: float):
        """ Posting data to receive vibration quality calculations from Brain"""
        return requests.post(f'{self.__brain_url}/vibration_by_params/', json={"group": group, "vibration": vibration})

    def get_oee_easy(self, oee_easy_input):
        """ Posting data to receive OEE value based on:
            Availability, Performance, Quality parameters from Brain """
        return requests.post(f'{self.__brain_url}/oee_easy/', json=oee_easy_input)

    def get_oee_full(self, oee_full_input):
        """ Posting data to receive oee value based on A, B, C, D, E, F parameters from Brain"""
        return requests.post(f'{self.__brain_url}/oee_full/', json=oee_full_input)

    def get_cp_indicator(self, cp_input):
        """
        Posting data to receive Cp indicator
        """

        return requests.post(f'{self.__brain_url}/cp_indicator/', json=cp_input)

    def get_pp_indicator(self, pp_input):
        """
        Posting data to receive Pp indicator
        """

        return requests.post(f'{self.__brain_url}/pp_indicator/', json=pp_input)

    def get_cpk_indicator(self, cpk_input):
        """
        Posting data to receive Cpk indicator
        """
        return requests.post(f'{self.__brain_url}/cpk_indicator/', json=cpk_input)

    def get_ppk_indicator(self, ppk_input):
        """
        Posting data to receive Ppk indicator
        """
        return requests.post(f'{self.__brain_url}/ppk_indicator/', json=ppk_input)

    def get_prediction(self, prediction_input):
        """
        Posting data to receive prediction value based on: module_name, variable_name, time_amount, time_unit,
        forecast_time_amount, forecast_unit parameters from Brain
        """
        return requests.post(f'{self.__brain_url}/prediction/', json=prediction_input)

    @staticmethod
    def parse_response(response):
        """ Parsing method"""
        if response.status_code == 200:
            json_response = response.json()
            if 'message' in json_response:
                print("Brain response failure")
                return None
            return json_response
        print("Brain response error")
        try:
            print(json.dumps(response.json(), indent=4))
        except:
            print("Code: ", response.status_code)
        return None



from nazca4sdk import UserVariableInfo, UserVariableDataType
from nazca4sdk.sdk import SDK

sdk = SDK()
#
# sdk.modules
# sdk.variables

print(f"Modules={sdk.modules}")
print(f"Varaibles={sdk.variables.list()}")
# print(sdk.user_variables.read([UserVariableInfo(UserVariableDataType.INT, "I1")], '2023-10-16T00:00:00', '2023-10-16T23:00:00'))
#print(sdk.read_variables_stats('symulator', ['P1', 'Q1'], '2023-01-24T00:00:00', '2023-01-24T23:00:00'))

print(sdk.cache.read_keys(["dupa"]))


"""Module to calculate energy quality according to EN50160"""
import logging
from typing import Optional

from pydantic import ValidationError
from nazca4sdk.analytics.__brain_connection import BrainClient


class Energy:
    """ Class to send and receive information from Energy module in brain"""

    def __init__(self):
        self.energy_brain = BrainClient()

    def calculate_energy_quality(self, module: str, npoe: Optional[bool] = False):
        """
        Function to determine energy quality values for determined input

        Args:
            npoe: Optional: if data should be import from npoe
            module: module name
        Returns:
            ::energy quality parameters -> dictionary with energy quality parameters::
            (worstCaseQuality: Overall energy quality;
            worstCaseQuality1: Overall energy quality of phase 1;
            worstCaseQuality2 Overall energy quality of phase 2;
            worstCaseQuality3: Overall energy quality of phase 3;
            frequencyQuality1: Overall frequency quality of phase 1;
            voltageQuality1: Overall voltage quality of phase 1;
            cosQuality1: Overall cosinus quality of phase 1;
            thdQuality1: Overall thd quality of phase 1;
            frequencyQuality2: Overall frequency quality of phase 2;
            voltageQuality2: Overall voltage quality of phase 2;
            cosQuality2: Overall cosinus quality of phase 2;
            thdQuality2: Overall thd quality of phase 2;
            frequencyQuality3: Overall frequency quality of phase 3;
            voltageQuality3: Overall voltage quality of phase 3;
            cosQuality3: Overall cosinus quality of phase 3;
            thdQuality3: Overall thd quality of phase 3;
        """

        try:
            response = self.energy_brain.get_energy_quality(module, npoe)
            result = self.energy_brain.parse_response(response)
            logging.info(f'Response: {result}')
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def calculate_energy_by_params(self, **kwargs):
        """
        Function to calculate energy quality by params
        Args:
            standard: Optional(int)
            freq1: Optional(float) frequency value phase 1;
            vol1: Optional(float) voltage value phase 1;
            cos1: Optional(float) cosinous value phase 1;
            thd1: Optional(float) thd value phase 1;
            freq2: Optional(float) frequency value phase 2;
            vol2: Optional(float) voltage value phase 2;
            cos2: Optional(float) cosinus value phase 2;
            thd2: Optional(float) thd value phase 2;
            freq3: Optional(float) frequency value phase 3;
            vol3: Optional(float) voltage value phase 3;
            cos3: Optional(float) cosinus value phase 3;
            thd3: Optional(float) thd value phase 3;

        Returns:
            ::energy quality parameters -> dictionary with energy quality parameters::
            (worstCaseQuality: Overall energy quality;
            worstCaseQuality1: Overall energy quality of phase 1;
            worstCaseQuality2 Overall energy quality of phase 2;
            worstCaseQuality3: Overall energy quality of phase 3;
            frequencyQuality1: Overall frequency quality of phase 1;
            voltageQuality1: Overall voltage quality of phase 1;
            cosQuality1: Overall cosinus quality of phase 1;
            thdQuality1: Overall thd quality of phase 1;
            frequencyQuality2: Overall frequency quality of phase 2;
            voltageQuality2: Overall voltage quality of phase 2;
            cosQuality2: Overall cosinus quality of phase 2;
            thdQuality2: Overall thd quality of phase 2;
            frequencyQuality3: Overall frequency quality of phase 3;
            voltageQuality3: Overall voltage quality of phase 3;
            cosQuality3: Overall cosinus quality of phase 3;
            thdQuality3: Overall thd quality of phase 3;
        """

        try:
            for i in kwargs.keys():
                if i not in ["freq1", "vol1", "cos1", "thd1", "freq2", "vol2", "cos2", "thd2", "freq3", "vol3", "cos3",
                             "thd3", "standard"]:
                    return NameError('Energy parameters should contains these names: freq1, vol1, cos1, thd1, freq2,'
                                     'vol2, cos2, thd2, freq3, vol3, cos3, thd3, standard')
            response = self.energy_brain.get_energy_quality_by_params(dict(**kwargs))
            result = self.energy_brain.parse_response(response)
            logging.info(f'Response: {result}')
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def calculate_optimal_contracted_capacity(self, start_date: str, end_date: str, connection: str,
                                              simulated_ordered_power: int = 0,
                                              ordered: bool = False, simulated: bool = False):
        """
        Function to calculate optimal contracted capacity for determined input

        Args:
            start_date: start date in UTC
            end_date: start date in UTC
            connection:
            simulated_ordered_power: simulated power by user, or 0 if power should be read from database
            ordered: bool value, if power should be read from database or from user
            simulated: bool value, if algorithm should simulate return values with provided simulatedOrderedPower
        Returns:
            ::optimal contracted capacity -> dictionary with optimal contracted capacity parameters::
            fee,
            penalty,
            power,
            summary
        """
        try:
            response = self.energy_brain.get_optimal_contracted_capacity(start_date,
                                                                         end_date,
                                                                         connection,
                                                                         simulated_ordered_power,
                                                                         ordered,
                                                                         simulated)
            result = self.energy_brain.parse_response(response)
            logging.info(f'Response: {result}')
            return result
        except ValidationError as error:
            print(error.json())
            return None



"""Module to communicating with file storage
"""

import os

from nazca4sdk.datahandling.open_data_client import OpenDataClient
from dependency_injector.wiring import inject


class FileStorage:
    """Communicate with file storage
    """

    @inject
    def __init__(self, opendata_client: OpenDataClient):
        self.opendata = opendata_client

    def download_file(self, path: str, local_path: str, override=False):
        """download file from file storage

        Args:
            path: path to file on file storage
            local_path: Path to save the file
            override: True - override existing file
        Returns:
            True - file download and saved , False - download error
        """

        if os.path.exists(local_path) and not override:
            print(f"File {local_path} already exist. Set override=True to override file.")
            return False
        if local_path.endswith("\\") or local_path.endswith("/"):
            print(f"local_path should be a file")
            return False
        return self.opendata.download_file(path, local_path)

    def send_file(self, path: str, local_path: str):
        """send file to file storage

        Args:
            path: path where file will be store
            local_path: file to send

        Returns:
            True when file store to file storage
            False when something bad happen
        """
        if not os.path.exists(local_path):
            print(f"file {local_path} not exist")
            return False
        file = open(local_path, 'rb')

        return self.opendata.send_file(path, file)



from pydantic import BaseModel, StrictStr, validator


class KeyValue(BaseModel):
    key: StrictStr
    value: object

    @validator('value')
    def validate_key(cls, valid_value):
        if isinstance(valid_value, str) or isinstance(valid_value, int) or isinstance(valid_value, float):
            return valid_value
        raise ValueError('value should be string, int or float')



from nazca4sdk import SDK
from nazca4sdk import UserVariable, UserVariableInfo, UserVariableDataType


u = UserVariable(UserVariableDataType.TEXT, "text", 13.2)
u1 = UserVariable(UserVariableDataType.DOUBLE, "double", 43.1)
u2 = UserVariable(UserVariableDataType.INT, "int", 127)
u3 = UserVariable(UserVariableDataType.BOOL, "bool", True)
u4 = UserVariable(UserVariableDataType.BOOL, "bool", False)
u7 = UserVariable(UserVariableDataType.INT, "int", 236)
u5 = UserVariable(UserVariableDataType.TEXT, "text", "Text1")
u6 = UserVariable(UserVariableDataType.TEXT, "tex", "Text2")
ud = UserVariable(UserVariableDataType.DATETIME, "datetime", "2022-10-04T15:05:01")

sdk = SDK(False)
# print(sdk.write_hotstorage_variables([u, u1, u2, u3, u4, u5, u6, ud]))
#
# va = UserVariableInfo(name="int", type=UserVariableDataType.INT)
# va2 = UserVariableInfo(name="text", type=UserVariableDataType.TEXT)
# res = sdk.read_hotstorage_variables("2022-10-04T00:40:02", "2022-10-05T17:00:00", [va, va2])
# print(len(res))
# for x in res:
#     print(x)
print(sdk.variables.read_variables_stats(module="symulator", variables=["V1", "V2"],
                                         start_date="2022-09-30T00:00:00", end_date="2022-10-20T00:00:00"))

v1 = UserVariableInfo(name="David", type=UserVariableDataType.INT)
v2 = UserVariableInfo(name="Ziutek", type=UserVariableDataType.INT)

print(sdk.variables.read_user_variables_stats(variables=[v1, v2],
                                              start_date="2022-09-15T00:00:00",
                                              end_date="2022-10-20T00:00:00"))



"""Module to calculate KPIs"""
from pandas import DataFrame
import pandas as pd
import numpy as np
from pydantic import ValidationError
from nazca4sdk.analytics.performance_indicators.oee_calculations import Oee
from nazca4sdk.analytics.performance_indicators.process_indicators import ProcessIndicators


class Indicators:
    """
    Indicators module with key performance indicators functions
    """

    def __init__(self):
        self.oee = Oee()
        self.process_indicators = ProcessIndicators()

    def get_oee_simple(self, availability: float, performance: float, quality: float):
        """The Overall Equipment Effectiveness (OEE)

        The Overall Equipment Effectiveness (OEE) is a proven way to monitor
        and improve process efficiency.
        it is considered as a diagnostic tool since it does not provide a solution to a given problem.
        OEE is a key process indicator that measures effectiveness and deviations
        from effective machine performance.

        OEE is calculated as:
            OEE = Availability x Performance x Quality
        Args:
            availability: float
            performance: float
            quality: float

        Return:
            OEE value: float

        Example:
            get_oee_simple(availability: 50, performance: 30, quality: 60)
        """
        try:
            data = {"availability": availability,
                    "performance": performance,
                    "quality": quality
                    }
            result = self.oee.calculate_oee_simple(data)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_oee_complex(self, a: float, b: float, c: float, d: float, e: float, f: float):
        """The Overall Equipment Effectiveness (OEE)

        The Overall Equipment Effectiveness (OEE) is a proven way to monitor
        and improve process efficiency.
        it is considered as a diagnostic tool since it does not provide a solution to a given problem.
        OEE is a key process indicator that measures effectiveness and deviations
        from effective machine performance.

        Args:
        OEE depends on parameters as follows:
            A = Total available time
            B = Run time
            C = Production capacity
            D = Actual production
            E = Production output (same as actual production)
            F = Actual good products (i.e. product output minus scraps)
        where,
        A and B define Availability,
        C and D define Performance,
        E and F define Quality

        OEE is calculated as:

        OEE = (B/A) x (D/C) x (F/E)

        Returns:
            OEE value: float

        Example:
            get_oee_complex(A: 50, B: 40, C: 60, D: 20, E: 100, F: 10)
        """
        try:
            data = {"A": a,
                    "B": b,
                    "C": c,
                    "D": d,
                    "E": e,
                    "F": f
                    }
            result = self.oee.calculate_oee_complete(data)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_availability(self, run_time: float, total_time: float):
        """Availability

            Takes into account availability/time loss, which includes all events
            related to unplanned stops.
            (e.g. equipment failures, material shortages) and planned stops (e.g. changeover times).
            Availability measures the proportion of time a machine or cell runs
            from the total theoretical available time.
            Calculated as:
                Availability = Run time/Total available time
            Args:
                ::input -> dictionary with oee parameters::
                run_time, Run time in hours
                total_time, Total run time in hours

            Returns:
                availability value: float

            Example:
                get_availability(run_time: 30, total_time: 50)
            """
        try:
            data = {"run_time": run_time,
                    "total_time": total_time,
                    }
            result = self.oee.calculate_availability(data)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_performance(self, actual_production: float, production_capacity: float):
        """Performance

            Takes into account performance/speed loss, which includes all the factors
            (e.g. slow cycles, small stops)
            that prevent the machine or cell to operate at maximum/optimal speed.
            It measures the proportion of produced units from the total number of possible
            produced units in a given run.

            Calculated as:
                Performance = Actual production/Production capacity

            Args:
                ::performance_input -> dictionary with performance parameters::
                actual_production, actual production
                production_capacity, production capacity

            Returns:
                performance value: float

            Example:
                get_performance(actual_production: 30, production_capacity: 50)
            """
        try:
            data = {"actual_production": actual_production,
                    "production_capacity": production_capacity,
                    }
            result = self.oee.calculate_availability(data)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_quality(self, actual_products: float, production_output: float):
        """Quality

            Takes into account quality loss, which includes all the factors
            (e.g. reworks, scraps, defects).
            that lead to defective units that do not meet the customer’s quality
            standards and specifications.
            Quality measures the proportion of non-defective units compared
            to the total units produced.
            Calculated as:
                Quality = Actual good products/Product output
            Args:
                ::quality_input -> dictionary with performance parameters::
                actual_products, Actual good products
                production_output, Production output

            Returns:
                quality value: float

            Example:
                get_quality(actual_products: 30, production_output: 50)
            """
        try:
            data = {"actual_products": actual_products,
                    "production_output": production_output,
                    }
            result = self.oee.calculate_quality(data)
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_cp_indicator(self, module: str, variable: str, start_date: str, end_date: str, period: int, subgroups: int, estimation_type='samples', **limits: dict):
        """
        The function to calculate Process Capability Indicator (Cp)

        Args:
            module: module name,
            variable: variable name,
            start_date: start of date range,
            end_date: end of date range,
            period: when estimation_type = 'samples', this is number of samples,
                for estimation_type = 'time' this is number of seconds
            subgroups: number of samples in subgroups,
            estimation_type: 'time' to estimate std using time offset or 'samples'
            to estimate using number of samples offset,
            limits: lsl, usl or both

        Return:
            Cp value

        """
        try:

            data = {"module": module,
                    "variable": variable,
                    "start_date": start_date,
                    "end_date": end_date,
                    "period": period,
                    "subgroups": subgroups,
                    "estimation_type": estimation_type
                    }
            keys = limits.keys()
            if "lsl" in keys and "usl" in keys:
                data['lsl'] = limits.get("lsl")
                data['usl'] = limits.get("usl")
            elif "lsl" in keys:
                data['lsl'] = limits.get("lsl")
            elif "usl" in keys:
                data['usl'] = limits.get("usl")

            result = self.process_indicators.get_cp_indicator(data)
            if result is None:
                return None
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_pp_indicator(self, module: str, variable: str, start_date: str, end_date: str, **limits: dict):
        """
        The function to calculate Process Performance Indicator (Pp)

        Args:
            module: module name,
            variable: variable name,
            start_date: start of date range,
            end_date: end of date range,
            limits: lsl, usl or both

        Return:
            Pp value

        """
        try:
            data = {"module": module,
                    "variable": variable,
                    "start_date": start_date,
                    "end_date": end_date,
                    }
            keys = limits.keys()
            if "lsl" in keys and "usl" in keys:
                data['lsl'] = limits.get("lsl")
                data['usl'] = limits.get("usl")
            elif "lsl" in keys:
                data['lsl'] = limits.get("lsl")
            elif "usl" in keys:
                data['usl'] = limits.get("usl")

            result = self.process_indicators.get_pp_indicator(data)
            if result is None:
                return None
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_cpk_indicator(self, module: str, variable: str, start_date: str, end_date: str, period: float, subgroups: int, estimation_type='samples', **limits: dict):
        """
        The function to calculate Process Capability Index (Cpk)

        Args:
            module: module name,
            variable: variable name,
            start_date: start of date range,
            end_date: end of date range,
            period: number of samples between subgroups,
            subgroups: number of samples in subgroups,
            estimation_type: 'time' to estimate std using time offset or 'samples'
            to estimate using number of samples offset,
            limits: lsl, usl or both

        Return:
            Cpk value: float
        """
        try:

            data = {"module": module,
                    "variable": variable,
                    "start_date": start_date,
                    "end_date": end_date,
                    "period": period,
                    "subgroups": subgroups,
                    "estimation_type": estimation_type
                    }

            keys = limits.keys()
            if "lsl" in keys and "usl" in keys:
                data['lsl'] = limits.get("lsl")
                data['usl'] = limits.get("usl")
            elif "lsl" in keys:
                data['lsl'] = limits.get("lsl")
            elif "usl" in keys:
                data['usl'] = limits.get("usl")
            result = self.process_indicators.get_cpk_indicator(data)
            if result is None:
                return None
            return result
        except ValidationError as error:
            print(error.json())
            return None

    def get_ppk_indicator(self, module: str, variable: str, start_date: str, end_date: str, **limits: dict):
        """
        The function to calculate Process Performance Index (Ppk)

        Args:
            module: module name,
            variable: variable name,
            start_date: start of date range,
            end_date: end of date range,
            limits: lsl, usl or both

        Return:
            Ppk value

        """
        try:
            data = {"module": module,
                    "variable": variable,
                    "start_date": start_date,
                    "end_date": end_date,
                    }
            keys = limits.keys()
            if "lsl" in keys and "usl" in keys:
                data['lsl'] = limits.get("lsl")
                data['usl'] = limits.get("usl")
            elif "lsl" in keys:
                data['lsl'] = limits.get("lsl")
            elif "usl" in keys:
                data['usl'] = limits.get("usl")

            result = self.process_indicators.get_ppk_indicator(data)
            if result is None:
                return None
            return result
        except ValidationError as error:
            print(error.json())
            return None



from datetime import datetime
from typing import List

from pydantic import BaseModel, StrictStr, validator

DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
user_variable_data_types = ["bool", "int", "double", "datetime", "text"]


class UserVariableValue(BaseModel):
    name: StrictStr
    type: StrictStr
    value: object
    dateTime: datetime

    @validator('value')
    def validate_key(cls, valid_value):
        if isinstance(valid_value, str) or isinstance(valid_value, int) or isinstance(valid_value, float) or isinstance(
                valid_value, bool) or isinstance(valid_value, datetime):
            return valid_value
        raise ValueError('value should be string, int or float')

    @validator('dateTime')
    def validate_datetime(cls, valid_datetime):
        if isinstance(valid_datetime, datetime):
            return valid_datetime
        if isinstance(valid_datetime, str):
            return datetime.strptime(valid_datetime)
        raise ValueError(f"{valid_datetime} is not valid datetime")

    @validator("type")
    def validate_type(cls, valid_type):
        if valid_type not in user_variable_data_types:
            raise ValueError(f"type should be one of -{user_variable_data_types}")
        return valid_type

    @validator("value")
    def validate_value(cls, valid_value, values):
        if 'type' not in values:
            raise ValueError("Can't validate value, bad type")
        data_type = values['type'].lower()
        if data_type == 'bool':
            if isinstance(valid_value, int):
                return bool(valid_value)
            if not isinstance(valid_value, bool):
                raise ValueError(f"Value {valid_value} is not bool")
            if not isinstance(valid_value, bool):
                raise ValueError(f"Value {valid_value} is not bool")
        if data_type == 'int' and ((isinstance(valid_value, bool) and isinstance(valid_value, int)) or (
                not isinstance(valid_value, bool) and not isinstance(valid_value, int))):
            raise ValueError(f"Value {valid_value} is not int")
        elif data_type == 'int' and isinstance(valid_value, int):
            if valid_value > 2147483647:
                raise ValueError(f"Value  {valid_value} is more then max int 2147483647")
            if valid_value < -2147483648:
                raise ValueError(f"Value  {valid_value} is less then min int -2147483648")
        if data_type == 'double' and (not isinstance(valid_value, float) and not isinstance(valid_value, int)):
            raise ValueError(f"Value {valid_value} is not float")
        elif data_type == 'double' and isinstance(valid_value, float):
            if valid_value > 3.4000002e38:
                raise ValueError(f"Value  {valid_value} is more then max double 3.4e38")
            if valid_value < -3.4000002e38:
                raise ValueError(f"Value  {valid_value} is less then min double -3.4e38")
        if data_type == 'datetime':
            if isinstance(valid_value, str):
                return UserVariableValue.valid_datetime(valid_value)
            if not isinstance(valid_value, datetime):
                raise ValueError(f"{valid_value} is not valid datetime")
            return valid_value
        if data_type == 'text' and not isinstance(valid_value, str):
            raise ValueError(f"value {valid_value} should be string")
        return valid_value

    @staticmethod
    def valid_datetime(valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            return datetime.fromisoformat(valid_datetime)
        except ValueError:
            raise ValueError(
                f'Bad datetime format({valid_datetime}). Format should be iso8601')


class UserVariablesValues(BaseModel):
    __root__: List[UserVariableValue]

    def variables(self):
        return self.__root__



from datetime import datetime
from pydantic import BaseModel, validator
from nazca4sdk.datahandling.hotstorage.user_variable_info import UserVariableInfo
DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
DATE_TIME_EXAMPLE = "yyyy-mm-ddTHH:MM:SS"


class UserVariablesStatsInfo(BaseModel):
    variables = [UserVariableInfo]
    startDate: str
    endDate: str

    @validator("endDate")
    def validate_end_date(cls, valid_datetime):
        """Validator to check if time amount is correct

               Args:
                   valid_datetime: str

               Returns:
                   valid_datetime: str
               """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                f'Bad datetime format. Example {DATE_TIME_EXAMPLE} ')

    @validator("startDate")
    def valid_start_date(cls, valid_datetime):
        """Validator to check if time amount is correct

        Args:
            valid_datetime: str

        Returns:
            valid_datetime: str
        """
        try:
            datetime.strptime(valid_datetime, DATE_TIME_FORMAT)
            return valid_datetime
        except ValueError:
            raise ValueError(
                f'Bad datetime format. Example {DATE_TIME_EXAMPLE}')


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '3.5.0'
PACKAGE_NAME = 'nazca4sdk'
AUTHOR = 'You'
AUTHOR_EMAIL = 'you@email.com'
# URL = 'https://github.com/you/your_package'


LICENSE = 'Apache License 2.0'
DESCRIPTION = 'Describe your package in one sentence'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"
with open('requirements.txt') as f:
    required = f.read().splitlines()

INSTALL_REQUIRES = required

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      # url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )


from nazca4sdk.datahandling.hotstorage.query import Query


class NpoeQueryHelper:
    @staticmethod
    def get_boot_profile_query(start_date, end_date, meter, profile_id):
        return Query() \
            .SELECT(f"SystemName as Meter,"
                    f"MeasureTime,"
                    f"MeasureTimeUtc,"
                    f"Pt,"
                    f"'{profile_id}' as ProfileId") \
            .FROM(f"npoe.electric_meters_raw emr") \
            .WHERE(f"MeasureTimeUtc >= '{start_date.strftime('%Y-%m-%dZ%H:%M:%S')}' "
                   f"AND "
                   f"MeasureTimeUtc < '{end_date.strftime('%Y-%m-%dZ%H:%M:%S')}' "
                   f"AND "
                   f"SystemName = '{meter}'")

    @staticmethod
    def get_pt_max_window_query(start_date, end_date):
        return Query() \
            .SELECT(f"pgams.Connection AS Connection, \
                                    toStartOfHour(pgams.MeasureTime) AS MeasureTime, \
                                    toStartOfHour(pgams.MeasureTimeUtc) AS MeasureTimeUtc, \
                                    last_value(pgams.Time_Inc) AS Time_Inc, \
                                    max(pgams.Pt_Avg_Sum) AS Pt_Window_Max") \
            .FROM(f"(SELECT \
                                        pgam.Connection AS Connection, \
                                        pgam.MeasureTime AS MeasureTime, \
                                        pgam.MeasureTimeUtc AS MeasureTimeUtc, \
                                        pgam.Time_Inc AS Time_Inc, \
                                        sum(Pt_Avg) AS Pt_Avg_Sum \
                                    FROM \
                                    ( \
                                        SELECT \
                                            db_connections.SystemName AS Connection, \
                                            db_meters.SystemName AS Meter, \
                                            toStartOfInterval(electric_meters_raw.MeasureTime,toIntervalMinute(15)) AS MeasureTime, \
                                            toStartOfInterval(electric_meters_raw.MeasureTimeUtc,toIntervalMinute(15)) AS MeasureTimeUtc, \
                                            last_value(electric_meters_raw.Time_Inc) AS Time_Inc, \
                                            avg(Pt) AS Pt_Avg \
                                        FROM npoe.electric_meters_raw AS electric_meters_raw \
                                        INNER JOIN npoe.db_meters AS db_meters ON electric_meters_raw.SystemName = db_meters.SystemName \
                                        INNER JOIN npoe.db_connections AS db_connections ON db_meters.ConnectionId = db_connections.Id \
                                        WHERE MeasureTimeUtc >= '{start_date.strftime('%Y-%m-%dT%H:%M:%S')}' \
                                        AND   MeasureTimeUtc < '{end_date.strftime('%Y-%m-%dT%H:%M:%S')}' \
                                        GROUP BY \
                                            Connection, \
                                            Meter, \
                                            MeasureTime, \
                                            MeasureTimeUtc \
                                        ORDER BY \
                                            Connection ASC, \
                                            Meter ASC, \
                                            MeasureTime ASC, \
                                            MeasureTimeUtc ASC \
                                    ) AS pgam \
                                    GROUP BY \
                                        Connection, \
                                        MeasureTime, \
                                        MeasureTimeUtc, \
                                        Time_Inc) as pgams"
                  ) \
            .GROUP_BY(f"Connection, MeasureTime, MeasureTimeUtc") \
            .ORDER_BY(f"Connection ASC, MeasureTime ASC, MeasureTimeUtc ASC")



"""SDK to communicate with Nazca4.0 system """
from dependency_injector.wiring import inject
from nazca4sdk.container import Container
from nazca4sdk.datahandling.hotstorage.clickhouse.clickhouse_client import ClickhouseClient
from nazca4sdk.datahandling.hotstorage.query import Query


class SDK:
    """SDK for Nazca4 system"""

    def __init__(self, https: bool):
        """ Initializing the system, checking connection and caching system configuration
        if https is required then https = True"""

        # nadpisuje konfiguracje
        Container.config.https.from_value(str(https))
        # print(self.c.clickhouse_client.get("Select * from nazca.test"))
        self._system_cache = Container.system_cache()
        #: :class:`nazca4sdk.analytics.analytics.Analytics` analytics functions
        self.analytics = Container.analytics()
        #: :class:`nazca4sdk.datahandling.nazcavariables.nazca_variables_storage.NazcaVariablesStorage`
        # read nazca variables
        self.nazca_variables = Container.nazca_variable_storage()
        #: :class:`nazca4sdk.system.variables.Variables` read device variables and variables statistics
        self.variables = Container.variables()
        #: :class:`nazca4sdk.system.user_variables.UserVariables` read and write user variables and user variables statistics
        self.user_variables = Container.user_variables()
        #: :class:`nazca4sdk.datahandling.knowledge.knowledge_storage` read, write and delete knowledge
        self.knowledge = Container.knowledge_storage()
        # :class:`nazca4sdk.datahandling.cache.cache_storage.CacheStorage` read and write cache value
        self.cache = Container.cache_storage()
        # :class:`nazca4sdk.system.npoe.Npoe` npoe functions
        self.npoe = Container.npoe()
        # :class:`nazca4sdk.datahandling.file.FileStorage` Communicate with files repository
        self.file_storage = Container.file_storage()
        self._clickhouse_client = Container.clickhouse_client()
        if not self._system_cache.load:
            print("Init SDK failed")

    def _get_modules(self):
        """Read all nazca4 modules

        Returns:
             (:obj:`list` of :obj:`str`) : list all modules
        """

        return self._system_cache.modules

    modules = property(_get_modules)

    def read_by_query(self, query: Query):
        """
        Reads data from hotstorage using SQL query.
        Args:
            query: <str>: SQL Query content.

        Returns: <dataframe>

        """
        query = query.strip(';')

        max_limit = 100
        if not query:
            raise ValueError("Query can't be empty.")

        if "LIMIT" not in query:
            limit = max_limit
            query = query.strip(';') + f" LIMIT {limit}"
        return self._clickhouse_client.get(query)


if __name__ == "nazca4sdk.sdk":
    container = Container()
    container.init_resources()



from nazca4sdk.sdk import SDK
#
# sdk = SDK(False)
#
# print(sdk.modules)

from nazca4sdk.datahandling.variable_verificator import EnergyInput, VibrationInput
#
# params = EnergyInput(freq1=50, vol1 = 230, cos1=1, thd1=0, freq2=50, vol2 = 230, cos2=1, thd2=0, freq3=50, vol3 = 230, cos3=1, thd3=0)
#
from nazca4sdk.analytics.energy.energy_quality import EnergyQuality
energy_quality = EnergyQuality()
print(energy_quality.calculate_energy_by_params(EnergyInput(freq1=50, vol1 = 230, cos1=1, thd1=0,
                                                            freq2=50, vol2 = 230, cos2=1, thd2=0,
                                                            freq3=50, vol3 = 230, cos3=1, thd3=0, standard=1)))
#
from nazca4sdk.datahandling.variable_verificator import VibrationInput
from nazca4sdk.analytics.vibration.vibration_quality import VibrationQuality
vibration_quality = VibrationQuality()
print(vibration_quality.calculate_vibration_quality_by_params(VibrationInput(group = 'G1r', vibration = 1)))

# print(energy_quality.calculate_energy_quality(module = 'symulator'))


from nazca4sdk.sdk import SDK
import datetime as dat
from datetime import datetime as dt

sdk = SDK(False)

# start_date = f'{dat.date.today()}T00:00:00Z'
# end_date = f'{dat.date.today()}T{dt.now().strftime("%H:%M:%S")}Z'
#
# print(start_date)
# print (end_date)

print(sdk.variables)
# #
print(sdk.daily_media_usage('SD5500_Fanuc', 'Zuzycie'))
# #
print(sdk.daily_energy_usage('Socomec_LiniaD1'))

# print(sdk.variable_over_time('symulator', ['Ea_dec'], 1, 'DAY'))

print(sdk.daily_media_usage('symulator', 'Ea_dec'))



from nazca4sdk.sdk import SDK

sdk = SDK()

print(sdk.modules)

print(sdk.variables)

# df = sdk.variable_over_time('BCM', ['vRmsy'], 10, 'MINUTE')
# df = sdk.variable_over_time('symulator', ['V1'], 3, 'HOUR')
# df = sdk.variable_over_day("symulator", ["Q2"], "2022-10-18T07:34:41", "2022-10-18T09:00:00")
# # df = sdk.variable_over_time("symulator", ["Q2"], 3, "HOUR")
# print(df)

result = sdk.read_historical_variable(module_name='symulator',
                                      variable_names=['I3'],
                                      start_date='2022-11-01T00:00:00',
                                      end_date='2022-12-17T00:00:00',
                                      page_size=1000)
if result:
    page = 1
    for a in result:
        print(f"page={page} ---> {a}")
        page = page + 1
else:
    print(f"Result empty")



from datetime import datetime, timedelta
from tzlocal import get_localzone
from typing import Optional, List

from pandas import DataFrame

from nazca4sdk.datahandling.hotstorage.clickhouse.clickhouse_client import ClickhouseClient
from nazca4sdk.datahandling.hotstorage.query import Query
from nazca4sdk.datahandling.kafka.kafka_sender import KafkaSender
from nazca4sdk.datahandling.npoe.connection import Connections
from nazca4sdk.datahandling.npoe.hot_storage_value import HotStorageValue
from nazca4sdk.datahandling.npoe.query_helper import NpoeQueryHelper
from nazca4sdk.datahandling.open_data_client import OpenDataClient
from nazca4sdk.datahandling.variable_verificator import VariableIntervalInfo, DATE_TIME_FORMAT, \
    VariableIntervalSubtractionInfo, ElectricMetersDataValidator, PgAvgDataValidator, SsvSopOrderedPowerDataValidator, \
    OrganizationCo2DataValidator, OzeConnectionInvertersDataValidator, OzeConnectionsForMetersDataValidator, \
    OzeInvertersSummaryDataValidator, MetersDataValidator

from nazca4sdk.system.system_cache import SystemCache
from nazca4sdk.tools.time import get_time_delta
from dependency_injector.wiring import inject


class Npoe:
    """ Class to get data from npoe

    """

    @inject
    def __init__(self, cache: SystemCache,
                 opendata_client: OpenDataClient,
                 click: ClickhouseClient,
                 broker: KafkaSender):
        self._system_cache = cache
        self._clickhouse_client = click
        self._openData = opendata_client
        self._broker = broker

    def get_electric_meters_data(self, system_name: List[str], variables=None, **time: dict) -> Optional[DataFrame]:
        """
        Gets electric meters data in specific time range

        Args:
            system_name: list of electric meter devices,
            variables: list of electric meter variables
            time: possible pairs - start_date, end_date or time_amount, time_unit

        Returns:
            DataFrame: values for selected variables and time range

        Examples:
            read(system_name:["module"], start_date="2023-03-21TOO:OO:00")
            read(system_name:["module"], time_amount=1,time_unit="MINUTE")
            read(system_name:["module"], time_amount=1, time_unit="MINUTE", variables=["U1","U2"])
        """
        if variables is None:
            variables = []
        keys = time.keys()

        columns_to_validate = {'system_name': system_name, 'variables': variables}
        columns_info = ElectricMetersDataValidator(**columns_to_validate)
        if "start_date" in keys and "end_date" in keys:
            time_to_validate = {'start_date': time.get("start_date"),
                                'end_date': time.get("end_date")}

            time_info = VariableIntervalInfo(**time_to_validate)
            return self._get_electric_meters_data(
                system_names=columns_info.system_name,
                start_date=datetime.strptime(time_info.start_date, DATE_TIME_FORMAT),
                end_date=datetime.strptime(time_info.end_date, DATE_TIME_FORMAT),
                variables=columns_info.variables)

        elif "time_amount" in keys and "time_unit" in keys:
            time_to_validate = {'time_amount': time.get("time_amount"),
                                'time_unit': time.get("time_unit")}

            variable_info = VariableIntervalSubtractionInfo(**time_to_validate)
            end_date = datetime.now()
            start_date = end_date - get_time_delta(variable_info.time_unit, variable_info.time_amount)

            return self._get_electric_meters_data(
                system_names=system_name,
                start_date=start_date,
                end_date=end_date,
                variables=variables)

        print("time should contains start_date and end_date or time_unit and time_amount")
        return None

    def _get_electric_meters_data(self, system_names, start_date, end_date, variables) -> Optional[DataFrame]:
        """
        Gets electric meters variables in specific time range by connection with open database

        Args:
            system_names - electric meter names,
            start_time - beginning of the time range,
            stop_time - ending of the time range,
            variables - list of electric meter variables

        Returns:
            DataFrame: values for selected variable and time range

        """
        if variables is None:
            variables = []
        try:
            if len(variables) == 0:
                query = Query() \
                    .SELECT("*") \
                    .FROM(f"npoe.electric_meters_raw") \
                    .WHERE(f"SystemName IN ({system_names}) "
                           f"AND MeasureTime >= '{start_date.strftime('%Y-%m-%dT%H:%M:%S')}' "
                           f"AND MeasureTime <= '{end_date.strftime('%Y-%m-%dT%H:%M:%S')}' ")
            else:
                selected_columns = "SystemName, MeasureTimeLong, MeasureTime, MeasureTimeUtc," + ','.join(variables)
                query = Query() \
                    .SELECT(selected_columns) \
                    .FROM(f"npoe.electric_meters_raw") \
                    .WHERE(f"SystemName IN ({system_names}) "
                           f"AND MeasureTime >= '{start_date.strftime('%Y-%m-%dT%H:%M:%S')}' "
                           f"AND MeasureTime <= '{end_date.strftime('%Y-%m-%dT%H:%M:%S')}' ")

            df = self._clickhouse_client.get(query)

            if df.size > 0:
                df = df.sort_values('MeasureTime').reset_index(drop=True)

            return df
        except ValueError:
            print("Error - Get variable data from Nazca4")
            return None

    def get_pgavg_data(self, connection: str, **time: dict) -> Optional[DataFrame]:
        """
        Gets power_guard_agg_hour values in specific time range

        Args:
            connection: list of electric meter devices,
            time: possible pair - start_date, end_date

        Returns:
            DataFrame: values for selected connection and time range
        """

        keys = time.keys()
        if "start_date" in keys and "end_date" in keys:
            if 'Z' in time.get("start_date"):
                datetime_start_obj = datetime.strptime(time.get("start_date"), "%Y-%m-%dT%H:%M:%S.%fZ")
                start_date = datetime_start_obj.strftime("%Y-%m-%dT%H:%M:%S")

                datetime_end_obj = datetime.strptime(time.get("end_date"), "%Y-%m-%dT%H:%M:%S.%fZ")
                end_date = datetime_end_obj.strftime("%Y-%m-%dT%H:%M:%S")

                columns_to_validate = {'connection': connection, 'start_date': start_date, 'end_date': end_date}
            else:
                columns_to_validate = {'connection': connection, 'start_date': time.get("start_date"),
                                       'end_date': time.get("end_date")}
            columns_info = PgAvgDataValidator(**columns_to_validate)

            return self._get_pgavg_data(
                connection=columns_info.connection,
                start_date=datetime.strptime(columns_info.start_date, DATE_TIME_FORMAT),
                end_date=datetime.strptime(columns_info.end_date, DATE_TIME_FORMAT))
        else:
            print("time should contains start_date and end_date")
            return None

    def _get_pgavg_data(self, connection, start_date, end_date) -> Optional[DataFrame]:
        """
        Gets power_guard_agg_hour values in specific time range by connection with open database

        Args:
            connection - SystemName from database,
            start_date - beginning of the date range,
            end_date - ending of the date range,

        Returns:
            DataFrame: values for selected time range
        """
        try:
            query = Query() \
                .SELECT("MeasureTimeUtc, Pt_Window_Max as value") \
                .FROM(f"npoe.power_guard_agg_hour") \
                .WHERE(f"Connection LIKE '{connection}' "
                       f"AND MeasureTimeUtc >= '{start_date.strftime('%Y-%m-%dT%H:%M:%S')}' "
                       f"AND MeasureTimeUtc <= '{end_date.strftime('%Y-%m-%dT%H:%M:%S')}'")
            df = self._clickhouse_client.get(query).sort_values('MeasureTimeUtc').reset_index(drop=True)
            return df
        except Exception:
            raise ValueError(f"Error - Get variable data from Nazca4")

    def get_ssv_sop_orderedpower(self, connection) -> Optional[DataFrame]:
        """
        Gets Ssv, Sop, OrderedPower values for specific connection from database

        Args:
            connection: list of electric meter devices,

        Returns:
            DataFrame: Ssv, Sop, OrderedPower values for selected connection
        """
        columns_to_validate = {'connection': connection}
        columns_info = SsvSopOrderedPowerDataValidator(**columns_to_validate)

        return self._get_ssv_sop_orderedpower(connection=columns_info.connection)

    def _get_ssv_sop_orderedpower(self, connection) -> Optional[DataFrame]:
        try:
            if len(connection) != 0:
                query = Query() \
                    .SELECT("SOP, SSV, OrderedPower") \
                    .FROM(f"npoe.db_connections") \
                    .WHERE(f"SystemName LIKE '{connection}'")

                df = self._clickhouse_client.get(query)
            return df
        except ValueError:
            print("Error - Get connection data from Nazca4")
            return None

    def calculate_pt_max_window(self):
        """
        Calculate max window value from electric meters connections power total 15min averages
        and save results to hotstorage
        """
        self._calculate_pt_max_window()

    def _calculate_pt_max_window(self):
        try:
            end_date = datetime.utcnow().replace(minute=0, second=0)
            start_date = end_date - timedelta(hours=1)
            query = NpoeQueryHelper().get_pt_max_window_query(start_date, end_date)

            df = self._clickhouse_client.get(query)

            data = []
            for i, row in df.iterrows():
                variables = [row.Connection, row.MeasureTime.strftime("%Y-%m-%dT%H:%M:%S"),
                             row.MeasureTimeUtc.strftime("%Y-%m-%dT%H:%M:%S"), row.Time_Inc, row.Pt_Window_Max]
                data.append(variables)

            self._broker.send_message(topic="dataflow.fct.clickhouse",
                                      key="sdk npoe",
                                      data=[HotStorageValue(table_name="npoe.power_guard_agg_hour", data=data).__dict__]
                                      )
            return df
        except Exception:
            print("Error - Calculate pt max window")
            return None

    def check_boot_profiles_to_record(self):
        """
        Check for boot profiles that should be recorded. Copies raw data to hotstorage and marks profiles as recorded
        in config database.
        """
        self._get_boot_profiles()

    def _get_boot_profiles(self):
        boot_profiles = self._openData.get_boot_profiles()
        if boot_profiles is not None:
            boot_profiles_to_record = list(filter(lambda x: x.recorded is False, boot_profiles))
            for boot_profile in boot_profiles_to_record:
                for meter in boot_profile.meters:
                    try:
                        if "." in boot_profile.startTime:
                            start_time = boot_profile.startTime.split(".")[0] + "Z"
                        else:
                            start_time = boot_profile.startTime

                        start_date = datetime.strptime(start_time, '%Y-%m-%dT%H:%M:%SZ')
                        start_date = start_date.replace(tzinfo=get_localzone())
                        end_date = start_date + timedelta(hours=1)
                        local_time_now = datetime.now(get_localzone())
                        if local_time_now > end_date:
                            query = NpoeQueryHelper.get_boot_profile_query(start_date, end_date, meter.systemName,
                                                                           boot_profile.id)
                            df = self._clickhouse_client.get(query)
                            data = []
                            for i, row in df.iterrows():
                                variables = [row.Meter, row.MeasureTime.strftime("%Y-%m-%dT%H:%M:%S"),
                                             row.MeasureTimeUtc.strftime("%Y-%m-%dT%H:%M:%S"), row.Pt,
                                             row.ProfileId]
                                data.append(variables)

                            if len(data) > 0:
                                self._broker.send_message(topic="dataflow.fct.clickhouse",
                                                          key="sdk npoe",
                                                          data=[
                                                              HotStorageValue(table_name="npoe.boot_profiles_recorded",
                                                                              data=data).__dict__]
                                                          )
                                self._openData.set_boot_profiles_as_recorded([boot_profile.id])
                    except Exception:
                        print("Error - Boot profile recording error")
                        return None

    def get_organization_co2(self, organization_id: str) -> Optional[DataFrame]:
        """
        Gets Dates and Values of CO2 coefficients for specific organization

        Args:
            organization_id: (str) organization_id,

        Returns:
            DataFrame: Dates and Values of CO2 coefficient for selected organization
        """
        try:
            columns_to_validate = {'organization_id': organization_id}
            columns_info = OrganizationCo2DataValidator(**columns_to_validate)
            return self._get_organization_co2(organization_id=columns_info.organization_id)
        except Exception:
            print("Error = get organization values of co2 coefficients")
            return None

    def _get_organization_co2(self, organization_id) -> Optional[DataFrame]:
        """
        Gets Dates and Values of CO2 coefficients for specific organization with sql query

        Args:
            organization_id: (str) organization_id,

        Returns:
            DataFrame: Dates and Values of CO2 coefficient for selected organization
        """
        try:
            query = Query() \
                .SELECT("Date, Value") \
                .FROM(f"npoe.db_co2_coefficients") \
                .WHERE(f"OrganizationId = '{organization_id}'")
            print(query)
            df = self._clickhouse_client.get(query).sort_values('Date').reset_index(drop=True)
            print(df)
            return df
        except Exception:
            raise ValueError(f"Error = get organization values of co2 coefficients")

    def get_oze_connection_inverters(self, connection_id: str) -> Optional[DataFrame]:
        """
        Gets inverter SystemName for specific connection

        Args:
            connection_id: (str) connection_id,

        Returns:
            DataFrame: SystemNames of inverters for specific connection
        """
        try:
            columns_to_validate = {'connection_id': connection_id}
            columns_info = OzeConnectionInvertersDataValidator(**columns_to_validate)
            return self._get_oze_connection_inverters(connection_id=columns_info.connection_id)
        except Exception:
            print("Error = SystemNames of inverters for specific connection")
            return None

    def _get_oze_connection_inverters(self, connection_id) -> Optional[DataFrame]:
        """
        Gets inverter SystemName for specific connection with sql query

        Args:
            connection_id: (str) connection_id,

        Returns:
            DataFrame: SystemNames of inverters for specific connection
        """
        try:
            query = Query() \
                .SELECT("DISTINCT (SystemName)") \
                .FROM(f"npoe.inverters_agg_hour") \
                .WHERE(f"SystemName in (Select SystemName from npoe.db_producers dp WHERE "
                       f"ConnectionId = '{connection_id}')")

            df = self._clickhouse_client.get(query)
            return df
        except Exception:
            raise ValueError(f"Error = SystemNames of inverters for specific connection")

    def get_oze_connections_for_meters(self, meters: list) -> Optional[DataFrame]:
        """
        Gets meter SystemName for specific connection which has oze

        Args:
            meters: (list) meters,

        Returns:
            DataFrame: SystemName of meter for specific Connection which has oze
        """
        try:
            columns_to_validate = {'meters': meters}
            columns_info = OzeConnectionsForMetersDataValidator(**columns_to_validate)
            return self._get_oze_connections_for_meters(meters=columns_info.meters)
        except Exception:
            print("Error = SystemNames of meters for specific connection which has oze")
            return None

    def _get_oze_connections_for_meters(self, meters) -> Optional[DataFrame]:
        """
        Gets meter SystemName for specific connection which has oze with sql query

        Args:
            meters: (list) meters,

        Returns:
            DataFrame: SystemName of meter for specific Connection which has oze
        """
        try:
            meters_param = "','".join(meters)
            query = Query() \
                .SELECT("SystemName, ConnectionId") \
                .FROM(f"npoe.db_meters") \
                .WHERE(f"SystemName in('{meters_param}') and ConnectionId in (SELECT DISTINCT ConnectionId "
                       f"from npoe.db_producers dp)")
            df = self._clickhouse_client.get(query)
            return df
        except Exception:
            raise ValueError(f"Error = SystemNames of meters for specific connection which has oze")

    def get_inverters_summary_data(self, inverters: list, from_date: str, to_date: str) -> Optional[DataFrame]:
        """
        Gets time, produced and given energy from oze for specific SystemNames (inverters)

        Args:
            inverters: (list) inverters,
            from_date: (str) beginning of the time range
            to_date: (str) ending of the time range

        Returns:
            DataFrame: time, produced and given energy from oze for specific SystemNames (inverters)
        """
        try:
            columns_to_validate = {'inverters': inverters, 'from_date': from_date, 'to_date': to_date}
            columns_info = OzeInvertersSummaryDataValidator(**columns_to_validate)
            return self._get_inverters_summary_data(inverters=columns_info.inverters, from_date=columns_info.from_date,
                                                    to_date=columns_info.to_date)
        except Exception:
            print("Error = SystemNames of meters for specific connection which has oze")
            return None

    def _get_inverters_summary_data(self, inverters: list, from_date: str, to_date: str) -> Optional[DataFrame]:
        """
        Gets time, produced and given energy from oze for specific SystemNames (inverters) with sql query

        Args:
            inverters: (list) inverters,
            from_date: (str) beginning of the time range
            to_date: (str) ending of the time range

        Returns:
            DataFrame: time, produced and given energy from oze for specific SystemNames (inverters)
        """
        try:
            inverters_param = "','".join(inverters)
            query = Query() \
                .SELECT("MeasureTimeUtc, sum(I_Ac_Energy_Wh_Diff) as produced, sum(Ea_minus_Diff) as given") \
                .FROM(f"npoe.inverters_agg_hour") \
                .WHERE(f"SystemName IN ('{inverters_param}') AND MeasureTimeUtc >= '{from_date}' AND MeasureTimeUtc < "
                       f"'{to_date}' GROUP BY MeasureTimeUtc ORDER BY MeasureTimeUtc")
            print("Query:", query)
            df = self._clickhouse_client.get(query)
            return df
        except Exception:
            raise ValueError(f"Error = SystemNames of meters for specific connection which has oze")

    def get_meters_data(self, meters: list, from_date: str, to_date: str) -> Optional[DataFrame]:
        """
        Gets hourly energy consumption for specific SystemNames

        Args:
            meters: (list) meters,
            from_date: (str) beginning of the time range
            to_date: (str) ending of the time range

        Returns:
            DataFrame: hourly energy consumption for specific SystemNames
        """
        try:
            columns_to_validate = {'meters': meters, 'from_date': from_date, 'to_date': to_date}
            columns_info = MetersDataValidator(**columns_to_validate)
            return self._get_meters_data(meters=columns_info.meters, from_date=columns_info.from_date,
                                         to_date=columns_info.to_date)
        except Exception:
            print("Error = Energy consumption for specific SystemNames")
            return None

    def _get_meters_data(self, meters: list, from_date: str, to_date: str) -> Optional[DataFrame]:
        """
        Gets hourly energy consumption for specific SystemNames with sql query

        Args:
            meters: (list) meters,
            from_date: (str) beginning of the time range
            to_date: (str) ending of the time range

        Returns:
            DataFrame: hourly energy consumption for specific SystemNames
        """
        try:
            meters_param = "','".join(meters)
            query = Query() \
                .SELECT("MeasureTimeUtc,SystemName,Ea_Diff") \
                .FROM(f"npoe.electric_meters_agg_hour") \
                .WHERE(f"MeasureTimeUtc >= '{from_date}' AND MeasureTimeUtc < '{to_date}' AND SystemName IN "
                       f"('{meters_param}') order by MeasureTimeUtc")
            df = self._clickhouse_client.get(query)
            return df
        except Exception:
            raise ValueError(f"Error = Energy consumption for specific SystemNames")

    def get_connections(self) -> Optional[List[Connections]]:
        """
        Gets connections with meters for all organizations

        Returns:
            DataFrame: list of connections

        """

        try:
            df = self._openData.get_connections()
            return df
        except ValueError:
            print("Error - Get connections")
            return None
